

META = """

class Meta(object):
    __base__ = object

    def __init__(self, *args):
        self.args = args

    def build(self, value):
        if isinstance(self.args[0], Meta):
            return self.args[0].build(value)
        return self.args[0](value)
"""

DATA_WRAPPER = """# -*- coding: utf-8 -*-


class DataWrapper(object):

    def __init__(self, instance):
        object.__setattr__(self, 'instance', instance)

    def __setattr__(self, name, value):
        object.__setattr__(object.__getattribute__(self, 'instance'), name, value)

    def __getattribute__(self, name):
        try:
            if name in ['is_valid']:
                return object.__getattribute__(self, name)
            return object.__getattribute__(self, 'instance').__getattribute__(name)
        except AttributeError as e:
            return None

    def __eq__(self, rhs):
        if rhs is None:
            return object.__getattribute__(self, 'instance') is None
        return rhs is not None and rhs.name is not None and self.name == rhs.name

    def __ne__(self, rhs):
        if rhs is None:
            return object.__getattribute__(self, 'instance') is not None
        return rhs is not None and rhs.name is None or self.name != rhs.name

    def __hash__(self):
        return hash(self.name)

    def is_valid(self):
        return object.__getattribute__(self, 'instance') is not None
"""

INTRUSIVE = """# -*- coding: utf-8 -*-


class IntrusivePtr(object):

    def __init__(self, instance):
        object.__setattr__(self, 'instance', instance)

    def __setattr__(self, name, value):
        object.__setattr__(object.__getattribute__(self, 'instance'), name, value)

    def __getattribute__(self, name):
        if name in ['is_valid']:
            return object.__getattribute__(self, name)
        return object.__getattribute__(self, 'instance').__getattribute__(name)

    def __eq__(self, rhs):
        if rhs is None:
            return
        self_instance = object.__getattribute__(self, 'instance')
        rhs_instance = object.__getattribute__(rhs, 'instance') if isinstance(rhs, IntrusivePtr) else rhs
        return self_instance.__eq__(rhs_instance)

    def __hash__(self):
        return id(object.__getattribute__(self, 'instance'))

    def is_valid(self):
        return object.__getattribute__(self, 'instance') is not None


def make_intrusive(class_name, *args):
    return IntrusivePtr(class_name(*args))
"""

SERIALIZER_XML = """# -*- coding: utf-8 -*-
from .DataWrapper import DataWrapper
from .IntrusivePtr import IntrusivePtr
import xml.etree.ElementTree as ET


class SerializerXml(object):
    def __init__(self, node):
        self.node = node

    def serialize(self, obj, key, default_value=None):
        if isinstance(obj, DataWrapper) and obj is not None:
            self.serialize_attr(obj.name, key, default_value)
        elif isinstance(obj, IntrusivePtr):
            child = self.add_child(key)
            child.serialize_attr(obj.get_type(), 'type', default_value)
            obj.serialize_xml(child)
        elif hasattr(obj, 'serialize_xml'):
            obj.serialize_xml(self.add_child(key))
        elif isinstance(obj, list):
            self.serialize_list(obj, key)
        elif isinstance(obj, dict):
            self.serialize_dict(obj, key)
        else:
            self.serialize_attr(obj, key, default_value)

    def serialize_attr(self, obj, key, default_value):
        if obj != default_value:
            if obj is True:
                self.node.attrib[key] = 'true'
            elif obj is False:
                if default_value is not None:
                    self.node.attrib[key] = 'false'
            else:
                self.node.attrib[key] = str(obj)

    def serialize_dict(self, obj, key):
        node = SerializerXml(self.node) if not key else self.add_child(key)
        for k, v in obj.items():
            item = node.add_child('pair')
            item.serialize_key(k)
            item.serialize_value(v)

    def serialize_list(self, obj, key):
        node = SerializerXml(self.node) if not key else self.add_child(key)
        for item in obj:
            node.add_child('item').serialize_list_item(item)

    def add_child(self, key):
        node = ET.SubElement(self.node, key) if key else self.node
        return SerializerXml(node)

    def serialize_key(self, k):
        self.serialize(k, 'key')

    def serialize_value(self, v):
        self.serialize(v, 'value')

    def serialize_list_item(self, obj):
        if isinstance(obj, DataWrapper):
            self.serialize_attr(obj.name, 'value', '')
        elif isinstance(obj, IntrusivePtr):
            self.node.tag = obj.get_type()
            if obj.is_valid():
                obj.serialize_xml(self)
        elif hasattr(obj, 'serialize_xml'):
            self.serialize(obj, '')
        elif isinstance(obj, list):
            self.serialize_list(obj, '')
        elif isinstance(obj, dict):
            self.serialize_dict(obj, '')
        else:
            self.serialize_attr(obj, 'value', None)
"""

DESERIALIZER_XML = """# -*- coding: utf-8 -*-
from .Meta import Meta
from .BaseEnum import BaseEnum
from .DataWrapper import DataWrapper
from .IntrusivePtr import IntrusivePtr, make_intrusive
from .Factory import Factory


class DeserializerXml(object):
    def __init__(self, node):
        self.node = node

    def deserialize(self, key, meta, default_value=None):
        node = self.get_child(key)
        if meta.__base__ == BaseEnum:
            value = self.deserialize_attr(key, str, '')
            return getattr(meta, value) if hasattr(meta, value) else default_value
        if isinstance(meta, Meta):
            if meta.args[0] == dict:
                return node.deserialize_dict('', meta)
            if meta.args[0] == list:
                return node.deserialize_list('', meta)
            if meta.args[0] == DataWrapper:
                from .DataStorage import DataStorage
                value = self.deserialize_attr(key, str, '')
                return getattr(DataStorage.shared(), 'get' + meta.args[1].TYPE)(value)
            if meta.args[0] == IntrusivePtr:
                if node is not None and node.node is not None:
                    obj = Factory.build(node.node.attrib['type'])
                    obj.deserialize_xml(DeserializerXml(node.node))
                    return obj
                return None
        if hasattr(meta, 'deserialize_xml'):
            obj = meta()
            obj.deserialize_xml(node)
            return obj
        return self.deserialize_attr(key, meta, default_value)

    def deserialize_attr(self, key, meta, default_value):
        value = None
        if key and self.node is not None:
            if key in self.node.attrib:
                value = self.node.attrib[key]
            else:
                value = default_value
        if meta == bool:
            if isinstance(value, bool):
                return value
            if value is not None:
                value = value.lower()
            return True if value in ['true', 'yes'] else False if value in ['false', 'no'] else default_value if default_value else False
        if meta == str and not value and not default_value:
            return ''
        try:
            return meta(value or default_value or 0)
        except ValueError as e:
            if meta == int:
                return int(float(value or default_value or 0))
            raise e

    def deserialize_dict(self, key, meta):
        node = DeserializerXml(self.node) if not key else self.get_child(key)
        if node.node is None:
            return {}
        result = {}
        for item in node.node:
            k = DeserializerXml(item).deserialize_key(meta.args[1])
            v = DeserializerXml(item).deserialize_value(meta.args[2])
            result[k] = v
        return result

    def deserialize_list(self, key, meta):
        if self.node is None:
            return []
        result = []
        for child in self.node:
            item = DeserializerXml(child).deserialize_list_item(meta.args[1])
            result.append(item)
        return result

    def get_child(self, key):
        return DeserializerXml(self.node.find(key) if (key and self.node is not None) else self.node)

    def deserialize_key(self, meta: Meta):
        return self.deserialize('key', meta)

    def deserialize_value(self, meta: Meta):
        return self.deserialize('value', meta)

    def deserialize_list_item(self, meta):
        if meta.__base__ == BaseEnum:
            value = self.deserialize_attr('value', str, '')
            return getattr(meta, value)
        elif hasattr(meta, 'serialize_xml'):
            obj = meta()
            obj.deserialize_xml(self)
            return obj
        elif isinstance(meta, list) or (isinstance(meta, Meta) and meta.args[0] == list):
            return self.deserialize_list('', meta)
        elif isinstance(meta, dict):
            return self.deserialize_dict('', meta)
        elif isinstance(meta, Meta):
            if meta.args[0] == DataWrapper:
                from .DataStorage import DataStorage
                value = self.deserialize_attr('value', str, '')
                return getattr(DataStorage.shared(), 'get' + meta.args[1].TYPE)(value)
            if meta.args[0] == IntrusivePtr:
                obj = Factory.build(self.node.tag)
                if obj and obj.is_valid():
                    obj.deserialize_xml(self)
                return obj
        else:
            return self.deserialize_attr('value', meta, None)
"""

SERIALIZER_JSON = """# -*- coding: utf-8 -*-
from .DataWrapper import DataWrapper
from .IntrusivePtr import IntrusivePtr


class SerializerJson(object):
    def __init__(self, json):
        self.json = json

    def serialize(self, obj, key, default_value=None):
        if isinstance(obj, DataWrapper):
            self.serialize_attr(obj.name, key, default_value)
        elif isinstance(obj, IntrusivePtr):
            child = self.add_child(key)
            child.serialize_attr(obj.get_type(), 'type', default_value)
            obj.serialize_json(child)
            pass
        elif hasattr(obj, 'serialize_json'):
            obj.serialize_json(self.add_child(key))
        elif isinstance(obj, list):
            self.serialize_list(obj, key)
        elif isinstance(obj, dict):
            self.serialize_dict(obj, key)
        else:
            self.serialize_attr(obj, key, default_value)

    def serialize_attr(self, obj, key, default_value):
        if obj != default_value:
            self.json[key] = obj

    def serialize_dict(self, obj, key):
        js = SerializerJson(self.json) if not key else self.add_child_array(key)
        for k, v in obj.items():
            item = js.add_array_item('')
            item.serialize_key(k)
            item.serialize_value(v)

    def serialize_list(self, obj, key):
        js = SerializerJson(self.json) if not key else self.add_child_array(key)
        for item in obj:
            js.serialize_list_item(item)

    def add_child(self, key):
        if key:
            self.json[key] = {}
            return SerializerJson(self.json[key])
        else:
            return self

    def add_child_array(self, key):
        self.json[key] = []
        return SerializerJson(self.json[key])

    def add_array_item(self, key):
        self.json.append({})
        return SerializerJson(self.json[-1])

    def serialize_key(self, k):
        self.serialize(k, 'key')

    def serialize_value(self, v):
        self.serialize(v, 'value')

    def serialize_list_item(self, obj):
        self.json.append(None)
        if isinstance(obj, DataWrapper):
            self.json[-1] = obj.name
        elif isinstance(obj, IntrusivePtr):
            self.json[-1] = {}
            self.json[-1]['type'] = obj.get_type() if obj.is_valid() else ''
            if obj.is_valid():
                obj.serialize_json(SerializerJson(self.json[-1]))
            pass
        elif hasattr(obj, 'serialize_json'):
            self.json[-1] = {}
            serializer = SerializerJson(self.json[-1])
            serializer.serialize(obj, '')
        elif isinstance(obj, list):
            self.json[-1] = []
            SerializerJson(self.json[-1]).serialize_list(obj, '')
        elif isinstance(obj, dict):
            self.serialize_dict(obj, '')
        else:
            self.json[-1] = obj
"""

DESERIALIZER_JSON = """# -*- coding: utf-8 -*-
from .Meta import Meta
from .BaseEnum import BaseEnum
from .DataWrapper import DataWrapper
from .IntrusivePtr import IntrusivePtr, make_intrusive
from .Factory import Factory


class DeserializerJson(object):
    def __init__(self, json):
        self.json = json

    def deserialize(self, key, meta, default_value=None):
        js = self.json if not key else self.json[key] if key in self.json else {}
        if meta.__base__ == BaseEnum:
            value = DeserializerJson(js).deserialize_attr('', str, '')
            return getattr(meta, value) if hasattr(meta, value) else default_value
        if isinstance(meta, Meta):
            if meta.args[0] == dict:
                return DeserializerJson(js).deserialize_dict('', meta)
            if meta.args[0] == list:
                return DeserializerJson(js).deserialize_list('', meta)
            if meta.args[0] == DataWrapper:
                from .DataStorage import DataStorage
                value = DeserializerJson(js).deserialize_attr('', str, '')
                return getattr(DataStorage.shared(), 'get' + meta.args[1].TYPE)(value)
            if meta.args[0] == IntrusivePtr:
                if js and 'type' in js:
                    obj = Factory.build(js['type'])
                    # obj = make_intrusive(meta.args[1])
                    obj.deserialize_json(DeserializerJson(js))
                    return obj
                return None
        if hasattr(meta, 'deserialize_json'):
            obj = meta()
            obj.deserialize_json(DeserializerJson(js))
            return obj
        return DeserializerJson(js).deserialize_attr('', meta, default_value)

    def deserialize_attr(self, key, meta, default_value):
        if key:
            if key in self.json:
                return self.json[key]
            else:
                return default_value
        if self.json.__class__ == meta:
            return meta(self.json)
        return meta(self.json or default_value)

    def deserialize_dict(self, key, meta):
        js = DeserializerJson(self.json) if not key else self.get_child_array(key)
        if not js.json:
            return {}
        assert isinstance(js.json, list)
        result = {}
        for item in js.json:
            k = DeserializerJson(item).deserialize_key(meta.args[1])
            v = DeserializerJson(item).deserialize_value(meta.args[2])
            result[k] = v
        return result

    def deserialize_list(self, key, meta):
        if not self.json:
            return []
        assert isinstance(self.json, list)
        result = []
        for js in self.json:
            item = DeserializerJson(js).deserialize('', meta.args[1])
            result.append(item)
        return result

    def get_child_array(self, key):
        return DeserializerJson(self.json[key] if key in self.json else [])

    def deserialize_key(self, meta: Meta):
        return self.deserialize('key', meta)

    def deserialize_value(self, meta: Meta):
        return self.deserialize('value', meta)
"""


from ..base.generator_data_storage_base import GeneratorDataStorageBase


class GeneratorDataStorage(GeneratorDataStorageBase):

    def __init__(self):
        GeneratorDataStorageBase.__init__(self)

    def get_shared_method_body(self):
        return '''
if(!DataStorage.__instance)
{
    DataStorage.__instance = new DataStorage();
}
return DataStorage.__instance;'''

    def get_pattern_getter(self):
        return '''
if (name in this.{map})
{{
    return this.{map}[name];
}}
return null;'''

    def get_initialize_function_json_body(self):
        return '''let json = JSON.parse(content);
            this.deserialize_json(json);
            this._loaded = true;'''



# -*- coding: utf-8 -*-
import xml.etree.ElementTree as ET
import json
from .common import *
from ..DeserializerJson import DeserializerJson
from ..DeserializerXml import DeserializerXml
from ..Meta import Meta
from ..SerializerJson import SerializerJson
from ..SerializerXml import SerializerXml


class AllTypesChildren(object):
    TYPE = "AllTypesChildren"
    __slots__ = ["value"]

    def __init__(self):
        self.value = 0

    def __hash__(self):
        return id(self)

    def __eq__(self, rhs):
        result = True
        result = result and self.value == rhs.value
        return result
        pass

    def __ne__(self, rhs):
        return not (self == rhs)
        pass

    def get_type(self):
        return AllTypesChildren.TYPE
        pass

    def serialize_xml(self, serializer: SerializerXml):
        pass

    def deserialize_xml(self, serializer: DeserializerXml):
        pass

    def serialize_json(self, serializer: SerializerJson):
        serializer.serialize(self.value, 'value', 0)
        pass

    def deserialize_json(self, deserializer: DeserializerJson):
        self.value = deserializer.deserialize('value', int, 0)




CMAKE = '''cmake_minimum_required(VERSION 3.13)
SET(ROOT ${CMAKE_SOURCE_DIR})
SET(GEN_DIR "@{gen_dir}")

option(WITH_DATA "Build with data" OFF)

project(@{project_name})

file(GLOB_RECURSE SRC ${GEN_DIR}/*.cpp ${ROOT}/external/*.cpp)

include_directories(${GEN_DIR} ${ROOT}/external)

if(WIN32)
    add_definitions(-W1 -std=c++17)
else()
    add_definitions(-Wall -std=c++17)
endif()

if(WITH_DATA)
    add_definitions(-DWITH_DATA=1)
endif(WITH_DATA)

'''

CMAKE_APP = '''add_executable(${PROJECT_NAME} ${SRC} ${ROOT}/main.cpp)
target_link_libraries(${PROJECT_NAME})'''

CMAKE_LIB = '''add_library(${PROJECT_NAME} STATIC ${SRC} ${ROOT}/mg.cpp)
target_link_libraries(${PROJECT_NAME})'''

MAIN_CPP_EXE = '''#include "Main.h"
#include "Registrar.h"

#ifdef WITH_DATA
#   include "DataStorage.h"
#   include <string>
#   include <fstream>
#   include <iostream>
#endif

int main(int argc, char ** args)
{
    mg::register_classes();

#ifdef WITH_DATA
    std::string path = "data/data.@{format}";
    std::fstream stream(path, std::fstream::in);
    if(!stream.is_open())
    {
        std::cout << "Cannot open file: " << path << std::endl;
        exit(-1);
    }
    std::string data((std::istreambuf_iterator<char>(stream)), std::istreambuf_iterator<char>());
    mg::DataStorage::shared().initialize_@{format}(data);
#endif

    mg::Main::main();
    return 0;
}
'''
MAIN_CPP_LIB = '''#include "Registrar.h"

#ifdef WITH_DATA
#   include "DataStorage.h"
#   include <string>
#   include <fstream>
#   include <iostream>
#endif

void initialize(int argc, char ** args)
{
    mg::register_classes();

#ifdef WITH_DATA
    std::string path = "data/data.@{format}";
    std::fstream stream(path, std::fstream::in);
    if(!stream.is_open())
    {
        std::cout << "Cannot open file: " << path << std::endl;
        exit(-1);
    }
    std::string data((std::istreambuf_iterator<char>(stream)), std::istreambuf_iterator<char>());
    mg::DataStorage::shared().initialize_@{format}(data);
#endif
}
'''
MAIN_HPP_LIB = '''#ifndef __mg_initialize_h__
#define __mg_initialize_h__

void initialize(int argc, char ** args);

#endif
'''

PROJECT_YAML = '''
project:
  {name}
  
binary_type:
  {binary_type}
'''

HELLO_WORLD_MLC = '''class Main
{
    function void main():static
    {
        print("Hello World!!!");
    }
}
'''


import unittest

import os
import sys
import inspect

sys.path.insert(0, os.path.abspath(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))) + '/..'))
from mlc_tools.utils.common import smart_split


class TestParseObject(unittest.TestCase):
    
    def test_1(self):
        parts = smart_split('int a, int b, int c', ',')
        self.assertEqual(len(parts), 3)
        self.assertEqual(parts[0], 'int a')
        self.assertEqual(parts[1], ' int b')
        self.assertEqual(parts[2], ' int c')


if __name__ == '__main__':
    unittest.main()


import re
from ..base.save_plugin_base import SavePluginBase


class SavePlugin(SavePluginBase):
    def __init__(self, model):
        SavePluginBase.__init__(self, model)

    def _create_combine_file_header(self):
        header = '''# -*- coding: utf-8 -*-
import json
import xml.etree.ElementTree as ET
'''
        file_name = 'mg.py'
        return header, file_name

    def _is_need_save_file_on_combine(self, file_name):
        return file_name == '__init__.py'

    def _remove_includes(self, file_content):
        file_content = re.sub(r'\n\s*from.+import.+', '', file_content)
        file_content = re.sub(r'\n\s*import.+', '', file_content)
        file_content = file_content.replace('# -*- coding: utf-8 -*-\n', '')
        return file_content



from ..base.serializer_base import SerializerBase, DESERIALIZATION
from ..core.object import Object, Objects
from .protocols import PHP_XML
from .protocols import PHP_JSON
from .regex import RegexPatternPhp


class Serializer(SerializerBase):

    def __init__(self):
        SerializerBase.__init__(self)

    def get_protocol(self, serialize_format):
        return PHP_XML if serialize_format == 'xml' else PHP_JSON

    def create_serialization_function(self, cls, serialize_type, serialize_format):
        method = SerializerBase.create_serialization_function(self, cls, serialize_type, serialize_format)

        use_data_storage = 'DataStorage::shared()' in method.body
        use_factory = 'Factory::shared()' in method.body
        imports = ''
        if serialize_type == DESERIALIZATION:
            if use_factory:
                imports += 'require_once "Factory.php";\n'
            if use_data_storage:
                imports += 'require_once "DataStorage.php";\n'
        if imports:
            method.body = imports + method.body
        return method.body

    def get_parent_serialize_format(self):
        return 'parent::{1}(${2});\n'

    def build_map_serialization(self, obj_name, obj_template_args, serialization_type, serialize_format):
        key = obj_template_args[0]
        value = obj_template_args[1]
        assert isinstance(key, Object) and isinstance(value, Object)
        key_type = key.type
        value_type = value.type
        pattern = self.serialize_protocol[serialization_type]['map'][0]
        value_is_pointer = value.is_pointer

        def get_create_type_operation(type_):
            types = {
                'list': 'array',
                'map': 'array',
            }
            return types[type_] if type_ in types else 'new ' + type_

        if value_type not in self.model.simple_types:
            value_declaration = '$value = {}();'.format(get_create_type_operation(value_type))
        else:
            value_declaration = ''
        key_serialize = self.build_serialize_operation_('key', key_type, None, serialization_type, key.template_args,
                                                        False, '$', key.is_link, serialize_format)
        value_serialize = self.build_serialize_operation_('value', value_type, None, serialization_type,
                                                          value.template_args,
                                                          value_is_pointer, '$', False, serialize_format)
        key_serialize = key_serialize.split('\n')
        for index, line in enumerate(key_serialize):
            key_serialize[index] = line
        key_serialize = '\n'.join(key_serialize)
        value_serialize = value_serialize.split('\n')
        for index, line in enumerate(value_serialize):
            value_serialize[index] = line
        value_serialize = '\n'.join(value_serialize)
        return pattern.format(field=obj_name,
                              key_serialize=key_serialize,
                              value_serialize=value_serialize,
                              key='{}',
                              owner='$this->',
                              value=value_declaration) + '\n'

    def convert_initialize_value(self, value):
        assert value is None or isinstance(value, str)

        if value in [None, 'nullptr', 'None']:
            value = 'null'
        if value and value.startswith('this'):
            value = '$' + value
        value = RegexPatternPhp.INITIALIZE[0].sub(RegexPatternPhp.INITIALIZE[1], value)
        return value

    def get_serialization_function_args(self, serialize_type, serialize_format):
        return ['xml', Objects.VOID] if serialize_format == 'xml' else ['json', Objects.VOID]

    def build_serialize_operation(self, obj, serialization_type, serialize_format):
        value = self.build_serialize_operation_(obj.name, obj.type, obj.initial_value,
                                                serialization_type, obj.template_args, obj.is_pointer, '$this->',
                                                obj.is_link, serialize_format)
        value = value.replace('(int64_t)', '(int)')
        value = value.replace('(uint)', '(int)')
        value = value.replace('(uint64_t)', 'int)')
        return value

    def finalize_serialize_operation(self, string):
        return string + '\n'



PHP_XML = '''
#string
#serialize:
#with default value:
if($(OWNER)$(FIELD) != $(DEFAULT_VALUE))
{
    $xml->addAttribute("$(FIELD)", $(OWNER)$(FIELD));
}
#without default value:
$xml->addAttribute("$(FIELD)", $(OWNER)$(FIELD));
#deserialize:
#with default value:
if($xml["$(FIELD)"])
{
    $(OWNER)$(FIELD) = (string)$xml["$(FIELD)"];
}
#without default value:
$(OWNER)$(FIELD) = (string)$xml["$(FIELD)"];


#float, int
#serialize:
#with default value:
if($(OWNER)$(FIELD) != $(DEFAULT_VALUE))
{
    $xml->addAttribute("$(FIELD)", $(OWNER)$(FIELD));
}
#without default value:
$xml->addAttribute("$(FIELD)", $(OWNER)$(FIELD));
#deserialize:
#with default value:
if($xml["$(FIELD)"])
{
    $(OWNER)$(FIELD) = ($(TYPE))$xml["$(FIELD)"];
}
#without default value:
$(OWNER)$(FIELD) = ($(TYPE))$xml["$(FIELD)"];

#bool
#serialize:
#with default value:
if($(OWNER)$(FIELD) != $(DEFAULT_VALUE))
{
    $xml->addAttribute("$(FIELD)", $(OWNER)$(FIELD) ? 'yes' : 'no');
}
#without default value:
$xml->addAttribute("$(FIELD)", $(OWNER)$(FIELD) ? 'yes' : 'no');
#deserialize:
#with default value:
if(isset($xml["$(FIELD)"]))
{
    $value = strtolower((string)$xml["$(FIELD)"]);
    $(OWNER)$(FIELD) = ($value == 'true' || $value == 'yes');
}
#without default value:
$value = strtolower((string)$xml["$(FIELD)"]);
$(OWNER)$(FIELD) = ($value == 'true' || $value == 'yes');


#serialized
#serialize:
$(OWNER)$(FIELD)->serialize_$(FORMAT)($xml->addChild("$(FIELD)"));
#deserialize:
$(OWNER)$(FIELD) = new $(TYPE);
$(OWNER)$(FIELD)->deserialize_$(FORMAT)($xml->$(FIELD));

#pointer
#serialize
if($(OWNER)$(FIELD))
{
    $child = $xml->addChild("$(FIELD)");
    $child->addAttribute("type", $(OWNER)$(FIELD)->get_type());
    $(OWNER)$(FIELD)->serialize_$(FORMAT)($child);
}
#deserialize:
if($xml->$(FIELD))
{
    $type = (string)$xml->$(FIELD)["type"];
    $(OWNER)$(FIELD) = Factory::build($type);
    if($(OWNER)$(FIELD))
    {
        $(OWNER)$(FIELD)->deserialize_$(FORMAT)($xml->$(FIELD));
    }
}


#list<int>, list<float>, list<string>
#serialize:
$xml_list = $xml->addChild("$(FIELD)");
foreach($(OWNER)$(FIELD) as $item)
{
    $xml_list->addChild("item")->addAttribute("value", $item);
}
#deserialize:
if($xml->$(FIELD))
{
    foreach($xml->$(FIELD)->item as $item)
    {
        array_push($(OWNER)$(FIELD), ($(ARG_0))$item["value"]);
    }
}
#list<bool>
#serialize:
$xml_list = $xml->addChild("$(FIELD)");
foreach($(OWNER)$(FIELD) as $item)
{
    $xml_list->addChild("item")->addAttribute("value", $item ? 'yes' : 'no');
}
#deserialize:
if($xml->$(FIELD))
{
    foreach($xml->$(FIELD)->item as $item)
    {
        $value = strtolower((string)$item["value"]);
        array_push($(OWNER)$(FIELD), ($value == 'true' || $value == 'yes'));
    }
}

#list<serialized>
#serialize:
$xml_list = $xml->addChild("$(FIELD)");
foreach($(OWNER)$(FIELD) as $item)
{
    $item->serialize_$(FORMAT)($xml_list->addChild("item"));
}
#deserialize:
if($xml->$(FIELD))
{
    foreach($xml->$(FIELD)->children() as $item)
    {
        $obj = new $(TYPE)();
        $obj->deserialize_$(FORMAT)($item);
        array_push($(OWNER)$(FIELD), $obj);
    }
}


#list<pointer>
#serialize:
$xml_list = $xml->addChild("$(FIELD)");
foreach($(OWNER)$(FIELD) as $item)
{
    $xml_child = $xml_list->addChild($item ? $item->get_type() : "null");
    if($item)
    {
        $item->serialize_$(FORMAT)($xml_list->addChild($item->get_type()));
    }
}
#deserialize:
if($xml->$(FIELD))
{
    foreach($xml->$(FIELD)->children() as $item)
    {
        $type = $item->getName();
        $obj = Factory::build($type);
        if($obj)
        {
            $obj->deserialize_$(FORMAT)($item);
        }
        array_push($(OWNER)$(FIELD), $obj);
    }
}


#link
#serialize:
$xml->addAttribute("$(FIELD)", is_string($(OWNER)$(FIELD)) ? $(OWNER)$(FIELD) : $(OWNER)$(FIELD)->name);
#deserialize:
$(OWNER)$(FIELD) = DataStorage::shared()->get$(TYPE)((string)$xml["$(FIELD)"]);


#list<link>
#serialize:
$xml_list = $xml->addChild("$(FIELD)");
foreach($(OWNER)$(FIELD) as $item)
{
    $xml_list->addChild("item")->addAttribute("value", is_string($item) ? $item : $item->name);
}
#deserialize:
foreach($xml->$(FIELD)->item as $item)
{
    array_push($(OWNER)$(FIELD), DataStorage::shared()->get$(ARG_0)((string)$item["value"]));
}

#map
#serialize:
$map_$(FIELD) = $xml->addChild("$(FIELD)");
$xml_main = $xml;
foreach($(OWNER)$(FIELD) as $key => $value)
{
    $xml = $map_$(FIELD)->addChild("pair");
    $(KEY_SERIALIZE)
    $(VALUE_SERIALIZE)
}
$xml = $xml_main;
#deserialize:
$xml_cache = $xml;
if($xml->$(FIELD))
{
    foreach($xml->$(FIELD)->children() as $xml_child)
    {
        $xml = $xml_child;
        $(VALUE)
        $(KEY_SERIALIZE)
        $map_key = (string)$key;
        $(VALUE_SERIALIZE)
        $(OWNER)$(FIELD)[$map_key] = $value;
    }
}
$xml = $xml_cache;


#enum
#serialize:
$xml->addAttribute("$(FIELD)", $(OWNER)$(FIELD));
#deserialize:
$(OWNER)$(FIELD) = (string)($xml["$(FIELD)"]);
'''

PHP_JSON = '''

#int,bool,float,string
#serialize:
#with default value:
if($(OWNER)$(FIELD) != $(DEFAULT_VALUE))
{
    $json->$(FIELD) = ($(TYPE))$(OWNER)$(FIELD);
}
#without default value:
$json->$(FIELD) = ($(TYPE))$(OWNER)$(FIELD);
#deserialize:
#with default value:
if(isset($json->$(FIELD)))
{
    $(OWNER)$(FIELD) = ($(TYPE))$json->$(FIELD);
}
#without default value:
$(OWNER)$(FIELD) = $json->$(FIELD);


#serialized
#serialize:
$json->$(FIELD) = json_decode("{}");
$(OWNER)$(FIELD)->serialize_$(FORMAT)($json->$(FIELD));
#deserialize:
if(isset($json->$(FIELD)))
{
    $(OWNER)$(FIELD) = new $(TYPE);
    $(OWNER)$(FIELD)->deserialize_$(FORMAT)($json->$(FIELD));
}
#pointer
#serialize
if($(OWNER)$(FIELD))
{
    $type = $(OWNER)$(FIELD)->get_type();
    $json->$(FIELD)        = json_decode("{}");
    $json->$(FIELD)->type  = $type;
    $(OWNER)$(FIELD)->serialize_$(FORMAT)($json->$(FIELD));
}
#deserialize:
if(isset($json->$(FIELD)))
{
    $type = (string) $json->$(FIELD)->type;
    $(OWNER)$(FIELD) = Factory::build($type);
    if($(OWNER)$(FIELD))
    {
        $(OWNER)$(FIELD)->deserialize_$(FORMAT)($json->$(FIELD));
    }
}


#list<int>, list<bool>, list<float>, list<string>
#serialize:
$json->$(FIELD) = array();
foreach($(OWNER)$(FIELD) as $item)
{
    array_push($json->$(FIELD), $item);
}
#deserialize:
if(isset($json->$(FIELD)))
{
    foreach($json->$(FIELD) as $item)
    {
        array_push($(OWNER)$(FIELD), $item);
    }
}


#list<serialized>
#serialize:
$json->$(FIELD) = array();
foreach($(OWNER)$(FIELD) as $item)
{
    array_push($json->$(FIELD), "");
    $item->serialize_$(FORMAT)($json->$(FIELD)[count($json->$(FIELD))]);
}
#deserialize:
foreach($json->$(FIELD) as $item)
{
    $obj = new $(TYPE)();
    if($obj)
    {
        $obj->deserialize_$(FORMAT)($item);
    }
    array_push($(OWNER)$(FIELD), $obj);
}

#list<pointer>
#serialize:
$json->$(FIELD) = array();
foreach($(OWNER)$(FIELD) as $t)
{
    $type       = $t ? $t->get_type() : "";
    $obj        = json_decode("{}");
    $obj->$type = json_decode("{}");
    if ($t)
    {
        $t->serialize_$(FORMAT)($obj->$type);
    }
    array_push($json->$(FIELD), $obj);
}
#deserialize:
$arr_$(FIELD) = $json->$(FIELD);
foreach($arr_$(FIELD) as $item)
{
    $type = key($item);
    $obj = Factory::build($type);
    if($obj)
    {
        $obj->deserialize_$(FORMAT)($item->$type);
    }
    array_push($(OWNER)$(FIELD), $obj);
}


#link
#serialize:
$json->$(FIELD) = $(OWNER)$(FIELD)->name;
#deserialize:
$name = $json->$(FIELD);
$(OWNER)$(FIELD) = DataStorage::shared()->get$(TYPE)((string)$name);

#list<link>
#serialize:
$json->$(FIELD) = array();
foreach($(OWNER)$(FIELD) as $data)
{
    array_push($json->$(FIELD), $data ? $data->name : "");
}
#deserialize:
foreach($json->$(FIELD) as $name)
{
    $data = DataStorage::shared()->get$(ARG_0)((string)$name);
    array_push($(OWNER)$(FIELD), $data);
}

#map
#serialize:
$map_$(FIELD) = array();
$json_main          = $json;
foreach ($(OWNER)$(FIELD) as $key => $value)
{
    $arr = json_decode("{}");
    $json = $arr;
    $(KEY_SERIALIZE)
    $(VALUE_SERIALIZE)

    array_push($map_$(FIELD), $arr);
}
$json = $json_main;
$json->$(FIELD) = $map_$(FIELD);
#deserialize:
$json_cache = $json;
if(isset($json->$(FIELD)))
{
    $arr = $json->$(FIELD);
    foreach($arr as $json)
    {
        $(VALUE)
        $(KEY_SERIALIZE)
        $map_key = (string)$key;
        $(VALUE_SERIALIZE)
        $(OWNER)$(FIELD)[$map_key] = $value;
    }
}
$json = $json_cache;



#enum
#serialize:
$json->$(FIELD) = $(OWNER)$(FIELD);
#deserialize:
if(isset($json->$(FIELD)))
{
    $(OWNER)$(FIELD) = (string)($json->$(FIELD));
}
'''





class TranslatorBase(object):

    def __init__(self):
        pass

    def translate(self, model):
        for cls in model.classes:
            for member in cls.members:
                if member.type == 'Observable':
                    member.is_runtime = True
                else:
                    for t in member.template_args:
                        if t.type == 'Observable':
                            member.is_runtime = True
                            break

            if cls.type == 'enum':
                self.convert_to_enum(cls)
            for method in cls.functions:
                self.translate_function(cls, method, model)

    def translate_function(self, cls, method, model):
        if not method.translated:
            body = '\n'.join(method.operations)
            body = self.translate_function_body(cls, method, body, model, method.args)
            method.body = body
        elif method.operations:
            method.body = '\n'.join(method.operations)

    def translate_function_body(self, cls, func, body, model, args):
        func = self.replace_by_regex(func, body, model, args)
        return func

    def replace_by_regex(self, func, body, model, args):
        pass

    def convert_to_enum(self, cls):
        shift = 0
        cast = 'string'
        values = []
        for member in cls.members:
            if member.name:
                continue
            member.name = member.type
            member.type = cast
            member.is_static = True
            member.is_const = True
            if member.initial_value is None:
                if cast == 'int':
                    member.initial_value = '(1 << {})'.format(shift)
                    values.append(1 << shift)
                elif cast == 'string':
                    member.initial_value = '"{}"'.format(member.name)
            else:
                # TODO if initialization is as enumerate of others members need throw error (example: one|two)
                cast = 'int'
                values.append(member.initial_value)

            shift += 1
        return values

    @staticmethod
    def replace(text, pattern):
        skip = len(pattern) > 2
        filters = pattern[2] if len(pattern) > 2 else []
        for filter_ in filters:
            skip = skip and text.find(filter_) == -1
        if not skip:
            text = pattern[0].sub(pattern[1], text)
        return text

    @staticmethod
    def save_strings(func):
        strings = []
        string_pattern = '@{__string_%d__}'
        while '"' in func:
            left = func.index('"')
            right = left + 1
            char = ''
            while right < len(func):
                if func[right] == '"' and char != '\\':
                    string = func[left:right + 1]
                    func = func[:left] + string_pattern % len(strings) + func[right + 1:]
                    strings.append(string)
                    break
                char = func[right]
                right += 1
        return func, strings

    @staticmethod
    def restore_strings(func, strings):
        string_pattern = '@{__string_%d__}'
        for i, string in enumerate(strings):
            func = func.replace(string_pattern % i, string)
        return func



# -*- coding: utf-8 -*-
import xml.etree.ElementTree as ET
import json

from .DataWrapper import DataWrapper
from .common import *
from .. import SerializerXml
from ..DeserializerXml import DeserializerXml
from ..SerializerJson import SerializerJson


class DataStorage(object):
    __instance = None
    TYPE = "DataStorage"

    def __init__(self):
        self.units = {}
        self._loaded = False

    def __hash__(self):
        return id(self)

    @staticmethod
    def shared():
        if not DataStorage.__instance:
            DataStorage.__instance = DataStorage()
        return DataStorage.__instance
        pass

    def initialize_json(self, content):
        js = json.loads(content)
        self.deserialize_json(js)
        self._loaded = True
        pass

    def getDataUnit(self, name):
        if not self._loaded and name not in self.units:
            from .DataUnit import DataUnit
            self.units[name] = DataUnit()
            self.units[name].name = name
        return DataWrapper(self.units[name])
        pass

    def get_type(self):
        return DataStorage.TYPE
        pass

    def serialize_xml(self, serializer: SerializerXml):
        pass

    def deserialize_xml(self, serializer: DeserializerXml):
        pass

    def serialize_json(self, serializer: SerializerJson):
        pass

    def deserialize_json(self, dictionary):
        pass



FACTORY = '''# -*- coding: utf-8 -*-
from .IntrusivePtr import make_intrusive

class Factory(object):

    @staticmethod
    def build(type):
{builders}
        return None
'''

COMMON_XML = '''

def create_command_from_xml(string):
    from .DeserializerXml import DeserializerXml
    import xml.etree.ElementTree as ET
    from .Factory import Factory
    root = ET.fromstring(string)
    type = root.tag
    command = Factory.build(type)
    if command is not None:
        deserializer = DeserializerXml(root)
        command.deserialize_xml(deserializer)
    return command


def serialize_command_to_xml(command):
    from .SerializerXml import SerializerXml
    import xml.etree.ElementTree as ET
    root = ET.Element(command.get_type())
    serializer = SerializerXml(root)
    command.serialize_xml(serializer)
    return ET.tostring(root)


def clone_object(obj, _=None):
    payload = serialize_command_to_xml(obj)
    clone = create_command_from_xml(payload)
    return clone
'''

COMMON_JSON = '''
def create_command_from_json(string):
    import json
    from .Factory import Factory
    from .DeserializerJson import DeserializerJson
    dictionary = json.loads(string)
    for key in dictionary:
        command = Factory.build(key)
        if command is not None:
            deserializer = DeserializerJson(dictionary[key])
            command.deserialize_json(deserializer)
        return command


def serialize_command_to_json(command):
    import json
    from .SerializerJson import SerializerJson
    js = dict()
    js[command.get_type()] = dict()
    serializer = SerializerJson(js[command.get_type()])
    command.serialize_json(serializer)
    return json.dumps(js)

def clone_object(obj, _=None):
    payload = serialize_command_to_json(obj)
    clone = create_command_from_json(payload)
    return clone
'''

MG_EXTENSIONS = '''

def strTo(value, class_):
    if class_ == bool:
        return value.lower() in ['yes', 'true', 'y']
    return class_(value)
    
def in_list(key, container):
    return key in container
    
def list_push(container, value):
    container.append(value)
    
def list_insert(container, value, index):
    container.insert(value, index)
    
def list_remove(container, value):
    container.remove(value)
    
def list_erase(container, index):
    container.remove(container[index])
    
def list_truncate(container, size):
    while len(container) > size:
        container.pop()
        
def list_clear(container):
    container.clear()
    
def list_size(container):
    return len(container)
    
def list_resize(container, size):
    list_truncate(container, size)
    while len(container) < size:
        container.append(None)

def in_map(key, container):
    return key in container
    
def map_size(container):
    return len(container)
    
def map_remove(container, value):
    container.pop(value, None)
    
def map_clear(container):
    container.clear()

def string_empty(container):
    return len(container) == 0
    
def string_size(container):
    return len(container)

def split(string, delimiter):
    return string.split(delimiter)

def join(values, delimiter):
    return delimiter.join(values)

'''



from ..core.class_ import Class


PREDEFINED = '''<?php
class Observable{
    private $listeners = array();
    private $objects = array();

    public function add($object, $callback) {
        array_push($this->listeners, $callback);
        array_push($this->objects, $object);
    }
    public function remove($object) {
        $index = array_search($object, $this->objects);
        if($index !== false) {
            array_splice($this->listeners, $index, 1);
            array_splice($this->objects, $index, 1);
        }
    }
    public function notify(...$arg){
        for ($i = 0; $i < count($this->listeners); ++$i) {
            $func = $this->listeners[$i];
            $obj = $this->objects[$i];
            $obj->$func(...$arg);
        }
    }
};
?>
'''


class GeneratorObserver(object):

    def __init__(self):
        pass

    @staticmethod
    def get_mock():
        cls = Class()
        cls.name = GeneratorObserver.get_observable_name()
        cls.type = 'class'
        cls.auto_generated = False
        return cls

    @staticmethod
    def get_observable_name():
        return 'Observable'

    @staticmethod
    def generate(model):
        text = PREDEFINED
        filename = GeneratorObserver.get_observable_name() + '.php'
        text = text.replace('@{namespace}', 'mg')
        text = text.replace('@{name}', GeneratorObserver.get_observable_name())
        model.add_file(None, filename, text)

        model.add_class(GeneratorObserver.get_mock())



from .parser import Parser
from ..core.object import Objects
from ..core.class_ import Class
from ..core.function import Function
from ..utils.error import Error


class GeneratorVisitor(object):

    def __init__(self):
        self.model = None
        self.base_visitor_classes = {}
        self.acceptors_interfaces = []
        self.support_override_methods = False

    def generate(self, model, support_override_methods):
        self.model = model
        self.support_override_methods = support_override_methods

        # find visitor and bases classes
        for cls in model.classes:
            base_class_name = self.get_base_visitor_name(cls)
            if base_class_name is None:
                continue
            if base_class_name not in self.base_visitor_classes:
                self.base_visitor_classes[base_class_name] = []
                base_class = model.get_class(base_class_name)
                self.base_visitor_classes[base_class_name].append(base_class)
            self.base_visitor_classes[base_class_name].append(cls)

        # generate acceptors interface, visit methods
        for base_class_name in self.base_visitor_classes:
            visitors = self.base_visitor_classes[base_class_name]
            self.generate_acceptor_interface(base_class_name, visitors)
            for visitor in visitors:
                self.add_accept_method(visitor, base_class_name)

        if not self.support_override_methods:
            # change name of methods to classes extends IVisitor interfaces
            for cls in model.classes:
                superclass_name = cls.superclasses[0] if cls.superclasses else None
                while superclass_name is not None:
                    if superclass_name in self.acceptors_interfaces:
                        self.override_methods(cls)
                        break
                    superclass = model.get_class(superclass_name)
                    superclass_name = superclass.superclasses[0] if superclass.superclasses else None

    def get_base_visitor_name(self, cls):
        for superclass_name in cls.superclasses:
            if superclass_name in self.base_visitor_classes:
                return superclass_name

            if not self.model.has_class(superclass_name):
                if superclass_name.startswith('IVisitor'):
                    # Correct situation. The class can be inherited from the IVisitor interface
                    continue
                Error.exit(Error.UNKNOWN_SUPERCLASS, cls.name, superclass_name)

            superclass = self.model.get_class(superclass_name)
            if superclass.is_visitor:
                return superclass_name

            result = self.get_base_visitor_name(superclass)
            if result:
                return result
        return None

    def generate_acceptor_interface(self, base_class_name, visitors):
        assert visitors
        acceptor = Class()
        acceptor.name = 'IVisitor' + base_class_name
        acceptor.group = visitors[0].group
        acceptor.type = "class"
        acceptor.is_abstract = True
        acceptor.is_virtual = True
        acceptor.side = visitors[0].side
        self.model.add_class(acceptor)
        self.acceptors_interfaces.append(acceptor.name)

        for visitor in visitors:
            method = Function()
            if self.support_override_methods:
                method.name = 'visit'
            else:
                method.name = 'visit_' + visitor.name[0].lower() + visitor.name[1:]
            method.return_type = Objects.VOID
            method.args.append(['ctx', Parser.create_object(visitor.name + '*')])
            method.is_abstract = True
            method.args[0][1].denied_intrusive = True
            acceptor.functions.append(method)

        if not self.support_override_methods:
            method = Function()
            acceptor.functions.append(method)
            method.name = 'visit'
            method.return_type = Objects.VOID
            method.args.append(['ctx', Parser.create_object(base_class_name + '*')])
            method.operations.append('''if(!ctx)
                {
                    return;
                }''')
            for visitor in visitors:
                method.operations.append('''else if(ctx->get_type() == {}::TYPE)
                {{
                    this->visit_{}(ctx);
                }}'''.format(visitor.name, visitor.name[0].lower() + visitor.name[1:]))

        def comparator(func):
            return func.args[0][1].type
        acceptor.functions.sort(key=comparator)

    @staticmethod
    def add_accept_method(class_, base_class_name):
        method = Function()
        method.name = 'accept'
        method.return_type = Objects.VOID
        method.args.append(['visitor', Parser.create_object('IVisitor' + base_class_name + '*')])
        method.operations.append('visitor->visit(this);')
        class_.functions.append(method)

    @staticmethod
    def override_methods(class_):
        for method in class_.functions:
            if method.name == 'visit' and len(method.args) == 1:
                arg_type = method.args[0][1].type
                method.name = 'visit_%s' % (arg_type[0].lower()) + arg_type[1:]



from ..utils.error import Error


class CircularReference(object):

    def __init__(self, model):
        self.model = model
        self.viewed = set()
        self.circular_reference = []

    def find(self):
        for cls in self.model.classes:
            self.viewed = set()
            self._wave(cls, cls)

        if self.circular_reference:
            self.circular_reference.append(self.circular_reference[0])
            Error.exit(Error.CIRCULAR_REFERENCE, '->'.join(self.circular_reference))

    def _wave(self, base_class, root_class):
        if root_class.name in self.viewed:
            return
        self.viewed.add(root_class.name)

        for member in root_class.members:
            if isinstance(member.type, str) and member.type == base_class.name and root_class.name != base_class.name:
                self.circular_reference.append(root_class.name)
            if self.model.has_class(member.type):
                cls = self.model.get_class(member.type)
                self._wave(base_class, cls)



from .generator_unit_tests_interface import GeneratorUnitTestsInterface


# pylint: disable=no-self-use
# pylint: disable=unused-argument
class GeneratorBase(object):

    def __init__(self):
        pass

    def generate(self, model):
        if model.generate_tests:
            GeneratorUnitTestsInterface().generate(model)



from ..core.class_ import Class
from ..core.object import Object, Objects
from ..core.object import AccessSpecifier
from ..core.function import Function
from ..base.parser import Parser
from ..base.model import SerializeFormat


def get_data_name(name):
    """
    convert DataNameFoo -> name_foo
    :param name: name of class in uppercase
    :return: name in lowercase
    """
    if name.find('Data') == 0:
        name = name[4:]
    name_ = ''
    for i, char in enumerate(name):
        if char.isupper() and i > 0:
            name_ += '_'
        name_ += char.lower()
    return name_


def get_class_name_from_data_name(name):
    """
    convert name_foo - > DataNameFoo
    :param name: name in lowercase
    :return: name of class in uppercase
    """
    upper = True
    name_ = ''
    for char in name:
        if char == '_':
            upper = True
            continue
        name_ += char if not upper else char.upper()
        upper = False
    return 'Data' + name_


def get_data_list_name(name):
    """ get plural form
    city -> cities
    unit -> units
    :param str name: name of unit
    :return:
    """
    last = name[-1]
    if last in 'y':
        if last in 'a,e,i,o,u,y':
            name = name[0:-1] + 'ies'
        else:
            name += 's'
    elif last in 'ou':
        name += 'es'
    elif last == 'f':
        name = name[0:-1] + 'ves'
    elif name[-2:-1] == 'fe':
        name = name[0:-2] + 'ves'
    elif last in ['s', 'ss', 'x', 'sh', 'ch']:
        name += 'es'
    else:
        name += 's'
    return name


class GeneratorDataStorageBase(Class):

    def __init__(self):
        Class.__init__(self)
        self.model = None
        self.data_members = {}

    def generate(self, model):
        self.model = model

        self.name = 'DataStorage'
        self.is_serialized = True

        for class_ in model.classes:
            if class_.is_storage and (class_.side in [model.side, 'both']):
                obj = Object()
                obj.type = 'map'
                obj.name = get_data_list_name(get_data_name(class_.name))
                obj.template_args.append(Objects.STRING)
                obj.template_args.append(class_.name)
                obj.access = AccessSpecifier.private
                self.members.append(obj)
                self.data_members[class_.name] = obj

                self.create_keys_getter(obj.name)

        loaded = Object()
        loaded.is_runtime = True
        loaded.name = '_loaded'
        loaded.type = 'bool'
        loaded.initial_value = 'false'
        loaded.access = AccessSpecifier.private
        self.members.append(loaded)

        self.create_shared_method()
        if model.serialize_formats & SerializeFormat.xml:
            self.add_initialize_function_xml()
        if model.serialize_formats & SerializeFormat.json:
            self.add_initialize_function_json()
        self.create_getters(model.classes)
        self.create_getters_maps(model.classes)
        model.add_class(self)

    def add_initialize_function_xml(self):
        method = Function()
        method.name = 'initialize_xml'
        method.return_type = Objects.VOID
        method.is_const = True
        method.args.append(['content', Objects.STRING])
        method.translated = True
        method.operations.append(self.get_initialize_function_xml_body())
        self.functions.append(method)

    def add_initialize_function_json(self):
        method = Function()
        method.name = 'initialize_json'
        method.return_type = Objects.VOID
        method.is_const = True
        method.args.append(['content', Objects.STRING])
        method.translated = True
        method.operations.append(self.get_initialize_function_json_body())
        self.functions.append(method)

    def get_initialize_function_json_body(self):
        pass

    def get_initialize_function_xml_body(self):
        pass

    def is_need_create_static_instance(self):
        return True

    def create_shared_method(self):
        if self.is_need_create_static_instance():
            obj = Object()
            obj.type = self.name
            obj.name = '__instance'
            obj.is_static = True
            obj.is_pointer = True
            obj.access = AccessSpecifier.private
            self.members.append(obj)

        method = Function()
        method.name = 'shared'
        method.return_type = Parser.create_object('{}&:const'.format(self.name))
        method.is_static = True
        method.translated = True
        method.operations.extend(self.get_shared_method_body().split('\n'))
        self.functions.append(method)

    def get_shared_method_body(self):
        assert False and 'Override me'

    def create_getters(self, classes):
        for class_ in classes:
            if class_.is_storage and class_.side in [self.model.side, 'both']:
                map_name = get_data_list_name(get_data_name(class_.name))
                method = Function()
                method.name = 'get' + class_.name
                method.args.append(['name', Objects.VOID])

                body = self.get_pattern_getter().format(map=map_name, type=class_.name)
                method.operations.append(body)
                method.translated = True
                self.functions.append(method)

    def get_pattern_getter(self):
        return ''

    def create_keys_getter(self, map_name):
        method = Function()
        method.name = f'get_{map_name}_keys'
        method.is_const = True
        method.return_type = Parser.create_object('list<string>')
        method.operations.append(f'''
        list<std::string> result;
        for(auto&& [key, _] : this->{map_name})
        {{
            list_push(result, key);   
        }}
        return result;
        ''')
        self.functions.append(method)

    def create_getters_maps(self, classes):
        for class_ in classes:
            if class_.is_storage and class_.side in [self.model.side, 'both']:
                method = Function()
                map_name = get_data_list_name(get_data_name(class_.name))
                method.name = f'get_{map_name}'
                method.is_const = True
                method.return_type = Parser.create_object(f'map<string, {class_.name}>:const:ref')
                method.operations.append(f'''
                return this->{map_name};
                ''')
                self.functions.append(method)



from .writer_base import WriterBase
from .generator import GeneratorBase
from .linker import Linker
from .parser import Parser
from .validator import Validator
from .data_parser import DataParser
from .language import Language
from .serializer_base import SerializerBase



from .constants import *
from ..base.writer_base import WriterBase
from ..base.model import Model


class GeneratorPredefinedFiles(object):

    def __init__(self):
        self.model = None

    def generate(self, model: Model):
        self.model = model
        writer = WriterBase('')
        writer.model = model

        for filename, content in FILES.items():
            model.add_file(None, filename, content)


FILES = {
    'common.php': COMMON
}


import re
from .regex import RegexPatternPhp
from ..base.translator_base import TranslatorBase


class Translator(TranslatorBase):

    def __init__(self):
        TranslatorBase.__init__(self)

    def replace_by_regex(self, func, body, model, args_):
        args = []
        args.extend(args_)
        args.extend(func.template_types)
        args = ', '.join(['$' + x[0] for x in args])

        if not body and not args:
            return body

        body, strings = self.save_strings(body)

        for reg in RegexPatternPhp.FUNCTION:
            body = self.replace(body, reg)

        for key in RegexPatternPhp.VARIABLES:
            patterns_dict = RegexPatternPhp.VARIABLES[key]
            arr = key.findall(args + '\n' + body)
            dividers = ' +-*\\=()[]<>\t\n!,.;'
            for var in arr:
                for char in dividers:
                    i = 0
                    while i < len(body):
                        replace = False
                        full = char + var
                        start = body.find(full, i)
                        if start == -1:
                            break
                        position = start + len(full)
                        i = position
                        if position < len(body):
                            replace = body[position] in dividers
                        if not replace:
                            continue
                        # func = func.replace(ch + var, ch + '$' + var)
                        body = body[:start] + (char + '$' + var) + body[position:]

                pattern = '^' + var
                if pattern not in patterns_dict:
                    patterns_dict[pattern] = re.compile(pattern)
                pattern = patterns_dict[pattern]

                body = pattern.sub('$' + var, body)
                for char in ['->']:
                    body = body.replace(char + '$' + var, char + var)

        for reg in RegexPatternPhp.FUNCTION_2:
            body = self.replace(body, reg)

        for reg in RegexPatternPhp.REPLACES:
            body = body.replace(reg[0], reg[1])

        body = self.restore_strings(body, strings)

        for cls in model.classes:
            if cls.name in body:
                body = 'require_once "{}.php";\n'.format(cls.name) + body

        return body



from .constants import FACTORY
from .writer import Writer


class GeneratorFactory(object):

    def __init__(self):
        pass

    @staticmethod
    def generate(model):
        writer = Writer('')
        writer.model = model
        line = '''
        if type == "{0}":
            from .{0} import {0}
            return make_intrusive({0})'''
        builders = ''
        for cls in model.classes:
            builders += line.format(cls.name)
        content = FACTORY.format(builders=builders)
        content = writer.prepare_file(content)
        model.add_file(None, 'Factory.py', content)



# -*- coding: utf-8 -*-
from tests.test_serializer.py.Meta import Meta
from tests.test_serializer.py.gen.BaseEnum import BaseEnum
from tests.test_serializer.py.gen.DataWrapper import DataWrapper
from tests.test_serializer.py.gen.IntrusivePtr import IntrusivePtr, make_intrusive
from tests.test_serializer.py.gen.Factory import Factory


class DeserializerJson(object):
    def __init__(self, json):
        self.json = json

    def deserialize(self, key, meta, default_value=None):
        js = self.json if not key else self.json[key] if key in self.json else {}
        if meta.__base__ == BaseEnum:
            value = DeserializerJson(js).deserialize_attr('', str, '')
            return getattr(meta, value) if hasattr(meta, value) else default_value
        if isinstance(meta, Meta):
            if meta.args[0] == dict:
                return DeserializerJson(js).deserialize_dict('', meta)
            if meta.args[0] == list:
                return DeserializerJson(js).deserialize_list('', meta)
            if meta.args[0] == DataWrapper:
                from tests.test_serializer.py.gen.DataStorage import DataStorage
                value = DeserializerJson(js).deserialize_attr('', str, '')
                return getattr(DataStorage.shared(), 'get' + meta.args[1].TYPE)(value)
            if meta.args[0] == IntrusivePtr:
                if js and 'type' in js:
                    obj = Factory.build(js['type'])
                    # obj = make_intrusive(meta.args[1])
                    obj.deserialize_json(DeserializerJson(js))
                    return obj
                return None
        if hasattr(meta, 'deserialize_json'):
            obj = meta()
            obj.deserialize_json(DeserializerJson(js))
            return obj
        return DeserializerJson(js).deserialize_attr('', meta, default_value)

    def deserialize_attr(self, key, meta, default_value):
        if key:
            if key in self.json:
                return self.json[key]
            else:
                return default_value
        if self.json.__class__ == meta:
            return meta(self.json)
        return self.json or default_value

    def deserialize_dict(self, key, meta):
        js = DeserializerJson(self.json) if not key else self.get_child_array(key)
        if not js.json:
            return {}
        assert isinstance(js.json, list)
        result = {}
        for item in js.json:
            k = DeserializerJson(item).deserialize_key(meta.args[1])
            v = DeserializerJson(item).deserialize_value(meta.args[2])
            result[k] = v
        return result

    def deserialize_list(self, key, meta):
        if not self.json:
            return []
        assert isinstance(self.json, list)
        result = []
        for js in self.json:
            item = DeserializerJson(js).deserialize('', meta.args[1])
            result.append(item)
        return result

    def get_child_array(self, key):
        return DeserializerJson(self.json[key] if key in self.json else [])

    def deserialize_key(self, meta: Meta):
        return self.deserialize('key', meta)

    def deserialize_value(self, meta: Meta):
        return self.deserialize('value', meta)



JS_JSON = '''
#int, bool, float, string
#serialize:
#with default value:
if($(OWNER)$(FIELD) != $(DEFAULT_VALUE))
{
    json["$(FIELD)"] = $(OWNER)$(FIELD);
}
#without default value:
json["$(FIELD)"] = $(OWNER)$(FIELD);
#deserialize
#with default value:
if("$(FIELD)" in json)
{
    $(OWNER)$(FIELD) = json["$(FIELD)"];
}
#without default value:
$(OWNER)$(FIELD) = json["$(FIELD)"];


#serialized
#serialize
if($(OWNER)$(FIELD))
{
    let dict = {};
    $(OWNER)$(FIELD).serialize_$(FORMAT)(dict);
    json["$(FIELD)"] = dict;
}
#deserialize
if('$(FIELD)' in json)
{
    $(OWNER)$(FIELD) = new $(TYPE)();
    $(OWNER)$(FIELD).deserialize_$(FORMAT)(json['$(FIELD)']);
}


#pointer
#serialize
if($(OWNER)$(FIELD))
{
    json['$(FIELD)'] = $({});
    json['$(FIELD)'].type = $(OWNER)$(FIELD).get_type();
    $(OWNER)$(FIELD).serialize_$(FORMAT)(json['$(FIELD)']);
}
#deserialize
if('$(FIELD)' in json)
{
    let type = json['$(FIELD)'].type;
    $(OWNER)$(FIELD) = Factory.build(type);
    $(OWNER)$(FIELD).deserialize_$(FORMAT)(json['$(FIELD)']);
}


#list<int>, list<float>, list<bool>, list<string>
#serialize
let arr_$(FIELD) = [];
for(let obj of $(OWNER)$(FIELD))
{
    arr_$(FIELD).push(obj);
}
json['$(FIELD)'] = arr_$(FIELD);
#deserialize
$(OWNER)$(FIELD) = [];
let arr_$(FIELD) = json['$(FIELD)']
for(let obj of arr_$(FIELD))
{
    $(OWNER)$(FIELD).push(obj)
}

#list<serialized>
#serialize
let arr_$(FIELD) = [];
for(let obj in $(OWNER)$(FIELD))
{
    let dict = {};
    obj.serialize_$(FORMAT)(dict);
    arr_$(FIELD).push(dict);
}
json['$(FIELD)'] = arr_$(FIELD);
#deserialize
$(OWNER)$(FIELD) = [];
let arr_$(FIELD) = json['$(FIELD)'];
for(let dict of arr_$(FIELD))
{
    let obj = new $(TYPE)();
    obj.deserialize_$(FORMAT)(dict);
    $(OWNER)$(FIELD).push(obj);
}


#list<pointer>
#serialize
json['$(FIELD)'] = [];
let arr_$(FIELD) = json['$(FIELD)'];
for(let t in $(OWNER)$(FIELD))
{
    arr_$(FIELD).push($({}));
    arr_$(FIELD)[-1][t.get_type()] = {};
    t.serialize_$(FORMAT)(arr_$(FIELD)[-1][t.get_type()]);
}
#deserialize
$(OWNER)$(FIELD) = [];
let arr_$(FIELD) = json['$(FIELD)'];
let size_$(FIELD) = arr_$(FIELD).length;
for(let index in size_$(FIELD))
{
    for(let key in arr_$(FIELD)[index])
    {
        let value = arr_$(FIELD)[index][key];
        let obj = Factory.build(key);
        $(OWNER)$(FIELD).push(obj);
        $(OWNER)$(FIELD)[-1].deserialize_$(FORMAT)(arr_$(FIELD)[index][key]);
        break;
    }
}


#link
#serialize:
if($(OWNER)$(FIELD))
{
    json['$(FIELD)'] = $(OWNER)$(FIELD).name;
}
#deserialize:
let name_$(FIELD) = json["$(FIELD)"];
if(name_$(FIELD))
{
    $(OWNER)$(FIELD) = DataStorage.shared().get$(TYPE)(name_$(FIELD));
}


#list<link>
#serialize:
json['$(FIELD)'] = [];
let arr_$(FIELD) = json['$(FIELD)'];
for(let t in $(OWNER)$(FIELD))
{
    arr_$(FIELD).push(t.name);
}
#deserialize:
$(OWNER)$(FIELD) = [];
let arr_$(FIELD) = json['$(FIELD)'];
for(let name in arr_$(FIELD))
{
    let data = DataStorage.shared().get$(ARG_0)(name);
    $(OWNER)$(FIELD).push(data);
}


#map
#serialize
let json_cache_$(FIELD) = json;
let arr_$(FIELD) = [];
json['$(FIELD)'] = arr_$(FIELD);
for(let key in $(OWNER)$(FIELD))
{
    let value = $(OWNER)$(FIELD)[key];
    arr_$(FIELD).push($({}));
    let json = arr_$(FIELD)[-1];
    $(KEY_SERIALIZE)
    $(VALUE_SERIALIZE)
}
json = json_cache_$(FIELD);
#deserialize
let json_cache_$(FIELD) = json;
let arr_$(FIELD) = json['$(FIELD)'];
for(let dict_key in arr_$(FIELD))
{
    let dict = arr_$(FIELD)[dict_key];
    let key = dict['key'];
    let type = key;
    let json = dict;
    let value = undefined;
    $(VALUE_SERIALIZE)
    $(OWNER)$(FIELD)[type] = value;
}
json = json_cache_$(FIELD);



#enum
#serialize:
json["$(FIELD)"] = $(OWNER)$(FIELD);
#deserialize
$(OWNER)$(FIELD) = json["$(FIELD)"];
'''





def add(var, filename):
    text = open(filename).read()
    text = text.replace('tests.test_serializer.py.gen', '')
    text = text.replace('tests.test_serializer.py', '')
    return '\n\n{} = """{}"""'.format(var, text)


content = ''
content += add('META', 'Meta.py')
content += add('DATA_WRAPPER', 'gen/DataWrapper.py')
content += add('INTRUSIVE', 'gen/IntrusivePtr.py')
content += add('SERIALIZER_XML', 'SerializerXml.py')
content += add('DESERIALIZER_XML', 'DeserializerXml.py')
content += add('SERIALIZER_JSON', 'SerializerJson.py')
content += add('DESERIALIZER_JSON', 'DeserializerJson.py')
open('../../../mlc_tools/module_python/constants_serializers.py', 'w').write(content)



from mlc_tools.core.class_ import Class
from mlc_tools.core.function import Function
from mlc_tools.core.object import Object, Objects, AccessSpecifier


class GeneratorRefCounter(object):

    def __init__(self):
        pass

    def generate(self, model):
        if not model.generate_ref_counter:
            return

        for cls in model.classes:
            if not cls.superclasses and cls.type != 'enum' and not cls.is_abstract:
                self._add(cls)

    @staticmethod
    def _add(cls: Class):
        if not cls.has_member_with_name('_reference_counter'):
            ref_counter = Object()
            ref_counter.name = '_reference_counter'
            ref_counter.type = 'int'
            ref_counter.initial_value = '1'
            ref_counter.is_runtime = True
            ref_counter.access = AccessSpecifier.private
            cls.members.append(ref_counter)
        if not cls.has_method_with_name('retain'):
            retain = Function()
            retain.name = 'retain'
            retain.return_type = Objects.VOID
            retain.operations.append('this->_reference_counter += 1;')
            cls.functions.append(retain)
        if not cls.has_method_with_name('release'):
            release = Function()
            release.name = 'release'
            release.return_type = Objects.INT
            release.operations.append('this->_reference_counter -= 1;')
            release.operations.append('auto counter = this->_reference_counter;')
            release.operations.append('if(counter == 0)')
            release.operations.append('delete this;')
            release.operations.append('return counter;')
            cls.functions.append(release)



# -*- coding: utf-8 -*-
import xml.etree.ElementTree as ET
import json
from .common import *
from ..DeserializerXml import DeserializerXml
from ..SerializerJson import SerializerJson
from ..SerializerXml import SerializerXml


class VisualUnit(object):
    TYPE = "VisualUnit"
    __slots__ = ["name", "icon"]

    def __init__(self):
        self.name = ""
        self.icon = ""

    def __hash__(self):
        return id(self)

    def __eq__(self, rhs):
        result = True
        result = result and self.name == rhs.name
        result = result and self.icon == rhs.icon
        return result
        pass

    def __ne__(self, rhs):
        return not (self == rhs)
        pass

    def get_type(self):
        return VisualUnit.TYPE
        pass

    def serialize_xml(self, serializer: SerializerXml):
        pass

    def deserialize_xml(self, serializer: DeserializerXml):
        pass

    def serialize_json(self, serializer: SerializerJson):
        pass

    def deserialize_json(self, dictionary):
        pass



import re
import xml.etree.ElementTree as ET

from tests.test_serializer.py.DeserializerJson import DeserializerJson
from tests.test_serializer.py.DeserializerXml import DeserializerXml
from tests.test_serializer.py.Meta import Meta
from tests.test_serializer.py.SerializerJson import SerializerJson
from tests.test_serializer.py.SerializerXml import SerializerXml
from tests.test_serializer.py.cpp_tests import JSON_TESTS, XML_TESTS
from tests.test_serializer.py.gen.AllTypes import AllTypes
from tests.test_serializer.py.gen.AllTypesChildren import AllTypesChildren
from tests.test_serializer.py.gen.DataStorage import DataStorage
from tests.test_serializer.py.gen.DataUnit import DataUnit
from tests.test_serializer.py.gen.DataWrapper import DataWrapper
from tests.test_serializer.py.gen.Point import Point
from tests.test_serializer.py.gen.TestEnum import TestEnum
from tests.test_serializer.py.gen.IntrusivePtr import make_intrusive, IntrusivePtr
from tests.test_serializer.py.gen.UnitType import UnitType
from tests.test_serializer.py.gen.common import serialize_command_to_json


def build_dict(parts, types, functor):
    if parts[0][0] in types:
        key = types[parts[0][0]]
    else:
        key = functor(parts[0][0])
    if parts[0][1] in types:
        value = types[parts[0][1]]
    else:
        value = functor(parts[0][1])
    return {key: value}


def get_object(description):
    parts = re.findall(r'map<(.+)\|(.+)>', description)
    types = {
        'int': 123,
        'bool': True,
        'float': 123.5,
        'std::string': '434312some_random',
        'TestEnum': TestEnum.value2,
        'const DataUnit*': DataStorage.shared().getDataUnit("unit1"),
        'AllTypesChildren': AllTypesChildren(),
        'intrusive_ptr<AllTypesChildren>': make_intrusive(AllTypesChildren),
        'std::vector<int>': [1, 2, 3, 4],
        'std::vector<float>': [1, 2],
        'std::vector<std::vector<bool>>': [[True, False], [False, True]],
        'std::map<int, int>': {1: 2, 2: 3},
        'std::vector<const DataUnit*>': [
            DataStorage.shared().getDataUnit("unit1"),
            DataStorage.shared().getDataUnit("unit1"),
        ],
        'std::vector<intrusive_ptr<AllTypesChildren>': [
            make_intrusive(AllTypesChildren),
            make_intrusive(AllTypesChildren),
        ],
        'std::vector<AllTypesChildren>': [
            AllTypesChildren(),
            AllTypesChildren()
        ]
    }
    return build_dict(parts, types, get_object)


def get_meta(description) -> Meta:
    parts = re.findall(r'map<(.+)\|(.+)>', description)
    types = {
        'int': int,
        'bool': bool,
        'float': float,
        'std::string': str,
        'TestEnum': TestEnum,
        'const DataUnit*': Meta(DataWrapper, DataUnit),
        'AllTypesChildren': AllTypesChildren,
        'intrusive_ptr<AllTypesChildren>': Meta(IntrusivePtr, AllTypesChildren),
        'std::vector<int>': Meta(list, int),
        'std::vector<float>': Meta(list, float),
        'std::vector<std::vector<bool>>': Meta(list, Meta(list, bool)),
        'std::map<int, int>': Meta(dict, int, int),
        'std::vector<const DataUnit*>': Meta(list, Meta(DataWrapper, DataUnit)),
        'std::vector<intrusive_ptr<AllTypesChildren>': Meta(list, Meta(IntrusivePtr, AllTypesChildren)),
        'std::vector<AllTypesChildren>': Meta(list, AllTypesChildren)
    }
    (k, v), = build_dict(parts, types, get_object).items()
    return Meta(dict, k, v)


def compare_json(counter, description, orig, serialized, total):
    if orig != serialized:
        counter += 1
        print('-------')
        print("Failed Json: ", description)
        print('Orig:', orig)
        print('Curr:', serialized)
        print('{}/{}'.format(counter, total))
        exit(1)
    return counter


def compare_xml(counter, description, orig, serialized, total):

    if ET.tostring(orig) != ET.tostring(serialized):
        counter += 1
        print('-------')
        print("Failed Xml: ", description)
        print('Orig:', ET.tostring(orig).decode())
        print('Curr:', ET.tostring(serialized).decode())
        print('{}/{}'.format(counter, total))
        exit(1)
    return counter


def test_json():
    counter = 0
    for description in JSON_TESTS:
        obj = get_object(description)
        meta = get_meta(description)
        orig = JSON_TESTS[description]

        try:
            serializer = SerializerJson({})
            serializer.serialize(obj, 'object')
            counter = compare_json(counter, description, orig, serializer.json, len(JSON_TESTS))

            deserializer = DeserializerJson(serializer.json)
            obj = deserializer.deserialize('object', meta)
            serializer = SerializerJson({})
            serializer.serialize(obj, 'object')
            counter = compare_json(counter, description, orig, serializer.json, len(JSON_TESTS))
        except AssertionError as e:
            print(e)
            counter = compare_json(counter, description, orig, {}, len(JSON_TESTS))
    print('Success Json: {}/{}'.format(len(JSON_TESTS) - counter, len(JSON_TESTS)))


def test_xml():
    counter = 0
    for description in XML_TESTS:
        orig = ET.fromstring(XML_TESTS[description])

        try:
            obj = get_object(description)
            meta = get_meta(description)
            serializer = SerializerXml(ET.Element('root'))
            serializer.serialize(obj, 'object')
            counter = compare_xml(counter, description, orig, serializer.node, len(XML_TESTS))

            deserializer = DeserializerXml(serializer.node)
            obj = deserializer.deserialize('object', meta)
            serializer = SerializerXml(ET.Element('root'))
            serializer.serialize(obj, 'object')
            counter = compare_xml(counter, description, orig, serializer.node, len(XML_TESTS))
        except AssertionError as e:
            print(e)
            counter = compare_xml(counter, description, orig, {})
        except IndexError as e:
            print(e)
            print('  - Cannot build object or meta for description: [%s]' % description)
            exit(1)
        # except TypeError as e:
        #     print('  -', e)
        #     print('  - Cannot save xml for description: [%s]' % description)
        #     exit(1)
    print('Success Xml: {}/{}'.format(len(XML_TESTS) - counter, len(XML_TESTS)))


def test_all_types_json():
    obj = AllTypes()
    obj.initialize()
    serializer = SerializerJson({})
    serializer.serialize(obj, 'object')
    orig = serializer.json

    deserializer = DeserializerJson(serializer.json)
    obj = deserializer.deserialize('object', AllTypes)
    serializer = SerializerJson({})
    serializer.serialize(obj, 'object')
    compare_json(0, 'AllTypes Json', orig, serializer.json, 1)
    print('Success AllTypes Json: {}/{}'.format(1, 1))


def test_all_types_xml():
    obj = AllTypes()
    obj.initialize()
    serializer = SerializerXml(ET.Element('root'))
    serializer.serialize(obj, 'object')
    orig = ET.fromstring(ET.tostring(serializer.node))

    deserializer = DeserializerXml(serializer.node)
    obj = deserializer.deserialize('object', AllTypes)
    serializer = SerializerXml(ET.Element('root'))
    serializer.serialize(obj, 'object')
    compare_xml(0, 'AllTypes Json', orig, serializer.node, 1)
    print('Success AllTypes Xml: {}/{}'.format(1, 1))


def test_deserialize_all_types_from_min_string_xml():
    deserializer = DeserializerXml(ET.fromstring('<object int_value0="5" />'))
    obj = AllTypes()
    obj.deserialize_xml(deserializer)
    assert obj.int_value0 == 5


def test_deserialize_all_types_from_min_string_json():
    deserializer = DeserializerJson({"int_value0": 10})
    obj = AllTypes()
    obj.deserialize_json(deserializer)
    assert obj.int_value0 == 10


def test_empty_enum_deserialize_xml():
    string = '<unit/>'
    deserializer = DeserializerXml(ET.fromstring(string))
    data = DataUnit()
    data.deserialize_xml(deserializer)
    assert data.unit_type == UnitType.defend
    assert isinstance(data.items, dict)


def test_empty_enum_deserialize_json():
    deserializer = DeserializerJson({})
    data = DataUnit()
    data.deserialize_json(deserializer)
    assert data.unit_type == UnitType.attack
    assert isinstance(data.items, dict)


def test_serialize_list_points_to_json():
    unit = DataUnit()
    unit.points.append(Point())
    unit.points[-1].row = 1
    unit.points[-1].col = 2
    js = serialize_command_to_json(unit)
    assert js == '{"DataUnit": {"points": [{"row": 1, "col": 2}]}}'



# -*- coding: utf-8 -*-
from tests.test_serializer.py.SerializerJson import SerializerJson
from tests.test_serializer.py.DeserializerJson import DeserializerJson


def create_command_from_json(string):
    import json
    from .Factory import Factory
    dictionary = json.loads(string)
    for key in dictionary:
        command = Factory.build(key)
        if command is not None:
            deserializer = DeserializerJson(dictionary[key])
            command.deserialize_json(deserializer)
        return command

def serialize_command_to_json(command):
    import json
    js = dict()
    js[command.get_type()] = dict()
    serializer = SerializerJson(js[command.get_type()])
    command.serialize_json(serializer)
    return json.dumps(js)

def clone_object(obj):
    payload = serialize_command_to_json(obj)
    clone = create_command_from_json(payload)
    return clone



from ..base.save_plugin_base import SavePluginBase


class SavePlugin(SavePluginBase):
    def __init__(self, model):
        SavePluginBase.__init__(self, model)

    def save_files(self, combine_to_one=True):
        SavePluginBase.save_files(self, False)



# -*- coding: utf-8 -*-
import xml.etree.ElementTree as ET
import json
from .common import *
from .UnitType import UnitType
from ..DeserializerJson import DeserializerJson
from ..DeserializerXml import DeserializerXml
from ..Meta import Meta
from ..SerializerJson import SerializerJson
from ..SerializerXml import SerializerXml


class DataUnit(object):
    TYPE = "DataUnit"
    __slots__ = ["name", "unit_type", "visual", "link_to_data", "all_units", "map_units", "items", 'points']

    def __init__(self):
        self.name = ""
        self.unit_type = UnitType()
        from .VisualUnit import VisualUnit
        self.visual = VisualUnit()
        self.link_to_data = None
        self.all_units = []
        self.map_units = {}
        self.items = {}
        self.points = []

    def __hash__(self):
        return id(self)

    def __eq__(self, rhs):
        result = True
        result = result and self.name == rhs.name
        result = result and self.unit_type == rhs.unit_type
        result = result and self.visual == rhs.visual
        result = result and self.all_units == rhs.all_units
        result = result and self.map_units == rhs.map_units
        return result
        pass

    def __ne__(self, rhs):
        return not (self == rhs)
        pass

    def get_type(self):
        return DataUnit.TYPE
        pass

    def serialize_xml(self, serializer: SerializerXml):
        pass

    def deserialize_xml(self, deserializer: DeserializerXml):
        self.unit_type = deserializer.deserialize('unit_type', UnitType, UnitType.defend)
        self.items = deserializer.deserialize("items", Meta(dict, str, int))

    def serialize_json(self, serializer: SerializerJson):
        serializer.serialize(self.points, "points")

    def deserialize_json(self, deserializer: DeserializerJson):
        self.unit_type = deserializer.deserialize('unit_type', UnitType, UnitType.attack)
        self.items = deserializer.deserialize("items", Meta(dict, str, int))



import os
import hashlib
import tempfile


def _get_files_list(path, prefix):
    try:
        files = os.listdir(path)
        result_files = []
        for i in files:
            if os.path.isdir(path + i):
                result = _get_files_list(path + i + '/', prefix + i + '/')
                for filename in result:
                    result_files.append(filename)
            if os.path.isfile(path + i):
                result_files.append(prefix + i)
        return result_files
    except IOError:
        return []
    except OSError as exception:
        print('Exception in [_get_files_list]:', exception, path, prefix)
        return []


def get_files_list(path):
    return _get_files_list(path, '')


def create_dir_for_file(filename):
    path = filename
    k = path.rindex('/')
    path = path[:k]
    if not os.path.exists(path):
        os.makedirs(path)


def remove(filename):
    if os.path.exists(filename):
        os.remove(filename)


def normalize_path(path, append_slash=True):
    path = path.replace('\\', '/')
    if append_slash and path and not path.endswith('/'):
        path += '/'
    return path


def load_dict(path):
    dictionary = {}
    try:
        file_ = open(path, encoding='utf-8')
        for line in file_:
            string = line.strip()
            args = string.split(' ')
            if len(args) == 2:
                key = string.split(' ')[0]
                value = string.split(' ')[1]
                dictionary[key] = value
        return dictionary
    except IOError:
        return dictionary


def save_dict(path, dict_):
    try:
        dictionary = load_dict(path)
        for key in dict_:
            dictionary[key] = dict_[key]
        with open(path, 'w', encoding='utf-8') as stream:
            for key in dictionary:
                string = key + ' ' + dictionary[key] + '\n'
                stream.write(string)
    except IOError:
        return False
    return True


def write(path, content):
    rewrite = True
    if os.path.exists(path):
        with open(path, encoding='utf-8') as stream:
            rewrite = stream.read() != content
    if rewrite:
        create_dir_for_file(path)
        with open(path, 'w', encoding='utf-8') as stream:
            if isinstance(content, bytes):
                content = content.decode()
            stream.write(content)
            return True, stream
    return False, None


__CACHE_FILES = tempfile.gettempdir() + '/bin/cache.tmp'


def file_has_changes(path):
    dictionary = load_dict(__CACHE_FILES)
    if path in dictionary:
        cache = dictionary[path]

        md5 = hashlib.md5()
        md5.update(open(path, encoding='utf-8').read())
        return not cache == str(md5.hexdigest())
    return True


def save_md5_to_cache(path):
    create_dir_for_file(__CACHE_FILES)
    md5 = hashlib.md5()
    md5.update(open(path, encoding='utf-8').read())
    save_dict(__CACHE_FILES, {path: md5.hexdigest()})



from ..base.generator_data_storage_base import GeneratorDataStorageBase
from ..base.generator_data_storage_base import get_data_list_name, get_data_name
from ..core.object import Object, AccessSpecifier, Objects
from ..core.function import Function


class GeneratorDataStorage(GeneratorDataStorageBase):

    def __init__(self):
        GeneratorDataStorageBase.__init__(self)

    def create_shared_method(self):
        GeneratorDataStorageBase.create_shared_method(self)

        obj = Object()
        obj.type = self.name
        obj.name = '__xml'
        obj.initial_value = 'NULL'
        obj.is_static = True
        obj.access = AccessSpecifier.private
        self.members.append(obj)

        obj = Object()
        obj.type = Objects.STRING
        obj.name = 'PATH_TO_DATA'
        obj.initial_value = '"assets/data/data.xml"'
        obj.is_static = True
        self.members.append(obj)

        obj = Object()
        obj.type = self.name
        obj.name = '__json'
        obj.initial_value = 'NULL'
        obj.is_static = True
        obj.access = AccessSpecifier.private
        self.members.append(obj)

        method = Function()
        method.name = 'deserialize_xml'
        method.args.append(['xml', Objects.VOID])
        self.functions.append(method)
        method = Function()
        method.name = 'deserialize_json'
        method.args.append(['xml', Objects.VOID])
        self.functions.append(method)

        method = Function()
        method.name = 'serialize_xml'
        method.args.append(['xml', Objects.VOID])
        self.functions.append(method)
        method = Function()
        method.name = 'serialize_json'
        method.args.append(['xml', Objects.VOID])
        self.functions.append(method)

    def get_shared_method_body(self):
        return SHARED_METHOD

    def create_getters(self, classes):
        for class_ in classes:
            if class_.is_storage and class_.side in [self.model.side, 'both']:
                map_name = get_data_list_name(get_data_name(class_.name))
                method = Function()
                method.name = 'get' + class_.name
                method.args.append(['name', Objects.VOID])
                method.operations.append(PATTERN_GETTER)
                method.operations[0] = method.operations[0].replace('@{array}', map_name)
                method.operations[0] = method.operations[0].replace('@{type}', class_.name)
                self.functions.append(method)

                method = Function()
                method.name = 'loadAll' + get_data_list_name(class_.name)
                method.operations.append(PATTERN_LOAD_ALL)
                method.operations[0] = method.operations[0].replace('@{array}', map_name)
                method.operations[0] = method.operations[0].replace('@{type}', class_.name)
                self.functions.append(method)

    def add_initialize_function_json(self):
        method = Function()
        method.name = 'initialize_json'
        method.return_type = Objects.VOID
        method.is_const = True
        method.args.append(['content', Objects.STRING])
        method.translated = True

        method.operations.append('$json = json_decode(content);')
        method.operations.append('$this->deserialize(json);')
        method.operations.append('$this->_loaded = true;')
        self.functions.append(method)

    def add_initialize_function_xml(self):
        method = Function()
        method.name = 'initialize_xml'
        method.return_type = Objects.VOID
        method.is_const = True
        method.args.append(['content', Objects.STRING])
        method.translated = True

        method.operations.append('$root = simplexml_load_string(content);')
        method.operations.append('$this->deserialize(root);')
        method.operations.append('$this->_loaded = true;')
        self.functions.append(method)


SHARED_METHOD = '''
if(DataStorage::$__instance == NULL)
{
    function endsWith($haystack, $needle)
    {
        //https://stackoverflow.com/questions/834303/startswith-and-endswith-functions-in-php
        // search forward starting from end minus needle length characters
        if ($needle === '')
        {
            return true;
        }
        $diff = \\strlen($haystack) - \\strlen($needle);
        return $diff >= 0 && strpos($haystack, $needle, $diff) !== false;
    }

    DataStorage::$__instance = new DataStorage();
    if(endsWith(DataStorage::$PATH_TO_DATA, '.xml'))
    {
        DataStorage::$__xml = simplexml_load_file(DataStorage::$PATH_TO_DATA);
    }
    else if(endsWith(DataStorage::$PATH_TO_DATA, '.json'))
    {
        $string = file_get_contents(DataStorage::$PATH_TO_DATA);
        DataStorage::$__json = json_decode($string);
    }

}
return DataStorage::$__instance;
'''

PATTERN_GETTER = '''
    if (array_key_exists($name, $this->@{array}))
    {
        return $this->@{array}[$name];
    }
    else if(DataStorage::$__xml != null)
    {
        $data = new @{type}();
        if(DataStorage::$__xml->@{array})
        {
            foreach (DataStorage::$__xml->@{array}->pair as $node)
            {
                if ($node["key"] == $name)
                {
                    $this->@{array}[$name] = $data;
                    $data->deserialize_xml($node->value);
                }
            }
        }
        return $data;
    }
    else if(DataStorage::$__json != null)
    {
        $data = new @{type}();
        if(DataStorage::$__json->@{array})
        {
            foreach (DataStorage::$__json->@{array} as $node)
            {
                if ($node->key == $name)
                {
                    $this->@{array}[$name] = $data;
                    $data->deserialize_json($node->value);
                }
            }
        }
        return $data;
    }
    '''

PATTERN_LOAD_ALL = '''
    if(DataStorage::$__xml != null && DataStorage::$__xml->@{array})
    {
        foreach (DataStorage::$__xml->@{array}->pair as $node)
        {
            $name = (string)$node["key"];
            $this->@{array}[$name] = new @{type}();
            $this->@{array}[$name]->deserialize_xml($node->value);
        }
    }
    else if(DataStorage::$__json != null && DataStorage::$__json->@{array})
    {
        foreach (DataStorage::$__json->@{array} as $node)
        {
            $name = (string)$node->key;
            $this->@{array}[$name] = new @{type}();
            $this->@{array}[$name]->deserialize_json($node->value);
        }
    }
    '''





class Modifiers(object):

    def __init__(self):
        pass

    abstract = ':abstract'
    serialized = ':serialized'
    static = ':static'
    const = ':const'
    external = ':external'
    visitor = ':visitor'
    set_function = ':set_function'
    runtime = ':runtime'
    key = ':key'
    link = ':link'
    storage = ':storage'
    private = ':private'
    protected = ':protected'
    public = ':public'
    numeric = ':numeric'
    test = ':test'
    inline = ':inline'
    virtual = ':virtual'
    pointer = ':pointer'
    ref = ':ref'
    generate = ':generate'
    prefer_use_forward_declarations = ':prefer_use_forward_declarations'

    client = ':client'
    server = ':server'
    side_client = 'client'
    side_server = 'server'




import unittest

import os
import sys
import inspect

sys.path.insert(0, os.path.abspath(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))) + '/..'))
from mlc_tools.module_cpp.translator import Translator as t_cpp
from mlc_tools.module_python.translator import Translator as t_py
from mlc_tools.module_php.translator import Translator as t_php
from mlc_tools.module_js.translator import Translator as t_js
from mlc_tools.core.function import Function
from mlc_tools.base.model import Model


class Test(unittest.TestCase):

    def test_simple(self):
        func = Function()
        model = Model()
        args = []

        body = 'auto foo = value ?? defaultValue'
        self.assertEqual(t_cpp().replace_by_regex(func, body, model, args), 'auto foo = (value != nullptr) ? value : (defaultValue)')
        self.assertEqual(t_py().replace_by_regex(func, body, model, args), 'foo = (value) if (value is not None) else (defaultValue)')
        self.assertEqual(t_php().replace_by_regex(func, body, model, args), '$foo = (value) ?? (defaultValue)')
        self.assertEqual(t_js().replace_by_regex(func, body, model, args), 'let foo = ((value) !== null && (value) !== undefined) ? (value) : (defaultValue)')

        body = 'Foo* foo = value->data.foo ?? defaultValue->data.foo'
        self.assertEqual(t_cpp().replace_by_regex(func, body, model, args), 'Foo* foo = (value->data.foo != nullptr) ? value->data.foo : (defaultValue->data.foo)')
        self.assertEqual(t_py().replace_by_regex(func, body, model, args), 'foo = (value.data.foo) if (value.data.foo is not None) else (defaultValue.data.foo)')
        self.assertEqual(t_php().replace_by_regex(func, body, model, args), '$foo = (value->data->foo) ?? (defaultValue->data->foo)')
        self.assertEqual(t_js().replace_by_regex(func, body, model, args), 'let foo = ((value.data.foo) !== null && (value.data.foo) !== undefined) ? (value.data.foo) : (defaultValue.data.foo)')


if __name__ == '__main__':
    unittest.main()



from ..base.translator_base import TranslatorBase
from ..core.object import Objects, Object
from ..core.function import Function
from .regex import RegexPatternCpp


class Translator(TranslatorBase):

    def __init__(self):
        TranslatorBase.__init__(self)

    def translate_function(self, cls, method, model):
        if not method.translated:
            body = '\n'.join(method.operations)
            body = self.translate_function_body(cls, method, body, model, method.args)
            method.body = body
        else:
            if method.operations:
                method.body = '\n'.join(method.operations)
            else:
                method.body = ''

    def translate_function_body(self, cls, func, body, model, args):
        if not body:
            body = ''
        body = self.replace_by_regex(func, body, model, args)
        return body

    def convert_to_enum(self, cls):
        cast = 'int'
        values = []

        def get_enum_value(cls_, index):
            return '(1 << {})'.format(index) if not cls_.is_numeric else str(index)

        def translate_members():
            shift = 0
            for member in cls.members:
                if member.name:
                    continue
                member.name = member.type
                member.type = cast
                member.is_static = True
                member.is_const = True
                if member.initial_value is None:
                    member.initial_value = get_enum_value(cls, shift)
                values.append(1 << shift)
                shift += 1

        def add_method(type_, name, const):
            method = Function()
            method.return_type = type_
            method.name = name
            method.is_const = const
            cls.functions.append(method)
            return method

        def create_const_ref():
            const_ref = Object()
            const_ref.type = cls.name
            const_ref.is_const = True
            const_ref.is_ref = True
            return const_ref

        def create_const_ref_base():
            const_ref = Object()
            const_ref.type = 'BaseEnum'
            const_ref.is_const = True
            const_ref.is_ref = True
            return const_ref

        def add_constructor_with_parameter():
            method = add_method(Object(), cls.name, False)
            method.args.append(['_value', Objects.INT])
            method.operations.append('value = _value;')

        def add_constructor_copy():
            method = add_method(Object(), cls.name, False)
            method.args.append(['rhs', create_const_ref()])
            method.operations.append('value = rhs.value;')

        def add_constructor_with_string():
            method = add_method(Object(), cls.name, False)
            method.args.append(['_value', Objects.STRING])
            for index, obj in enumerate(cls.members):
                if obj.name != 'value' and index <= len(values):
                    method.operations.append('''if(_value == "{0}")
                    {{
                        value = {0};
                        return;
                    }}'''.format(obj.name))
            method.operations.append('value = 0;')

        def add_operator_copy():
            method = add_method(create_const_ref(), 'operator =', False)
            method.args.append(['rhs', create_const_ref()])
            method.operations.extend(['value = rhs.value;', 'return *this;'])

        def add_operator_copy_with_int():
            method = add_method(create_const_ref(), 'operator =', False)
            method.args.append(['rhs', Objects.INT])
            method.operations.extend(['value = rhs;', 'return *this;'])

        def add_operator_copy_with_string():
            method = add_method(create_const_ref(), 'operator =', False)
            method.args.append(['_value', Objects.STRING])
            for index, obj in enumerate(cls.members):
                if obj.name != 'value' and index <= len(values):
                    method.operations.append('''if(_value == "{0}")
                                {{
                                    value = {0};
                                    return *this;
                                }}'''.format(obj.name))
            method.operations.append('return *this;')

        def add_operator_equals():
            method = add_method(Objects.BOOL, 'operator ==', True)
            method.args.append(['rhs', create_const_ref()])
            method.operations.append('return value == rhs.value;')

            method = add_method(Objects.BOOL, 'operator ==', True)
            method.args.append(['rhs', create_const_ref_base()])
            method.operations.append('return value == rhs.operator int();')

        def add_operator_equals_with_int():
            method = add_method(Objects.BOOL, 'operator ==', True)
            method.args.append(['rhs', Objects.INT])
            method.operations.append('return value == rhs;')

        def add_operator_equals_with_string():
            method = add_method(Objects.BOOL, 'operator ==', True)
            method.args.append(['rhs', Objects.STRING])
            method.operations.append('return *this == {}(rhs);'.format(cls.name))

        def add_friend_equals_with_string():
            method = add_method(Objects.BOOL, 'operator ==', False)
            method.is_friend = True
            method.args.append(['lhs', Objects.STRING])
            method.args.append(['rhs', create_const_ref()])
            method.operations.append('return {}(lhs) == rhs;'.format(cls.name))

        def add_operator_less():
            method = add_method(Objects.BOOL, 'operator <', True)
            method.args.append(['rhs', create_const_ref()])
            method.operations.append('return value < rhs.value;')

        def add_cast_to_int():
            method = add_method(Object(), 'operator int', True)
            method.operations.append('return value;')

        def add_operator_cast_string():
            method = add_method(Object(), 'operator std::string', True)
            for index, obj in enumerate(cls.members):
                if obj.name != 'value' and index <= len(values):
                    method.operations.append('''if(value == {0})
                    {{
                        return "{0}";
                    }}'''.format(obj.name))
            method.operations.append('return std::string();')

        def add_method_str():
            method = add_method(Objects.STRING, 'str', True)
            for index, obj in enumerate(cls.members):
                if obj.name != 'value' and index <= len(values):
                    method.operations.append('''if(value == {0})
                                {{
                                    return "{0}";
                                }}'''.format(obj.name))
            method.operations.append('return std::string();')

        translate_members()
        add_constructor_with_parameter()
        add_constructor_copy()
        add_constructor_with_string()
        add_operator_copy()
        add_operator_copy_with_int()
        add_operator_copy_with_string()
        add_operator_equals()
        add_operator_equals_with_int()
        add_operator_equals_with_string()
        add_friend_equals_with_string()
        add_operator_less()
        add_cast_to_int()
        add_operator_cast_string()
        add_method_str()

        return values

    def replace_by_regex(self, func, body, model, args):
        for reg in RegexPatternCpp.FUNCTION:
            body = self.replace(body, reg)
        for reg in RegexPatternCpp.REPLACES:
            body = body.replace(reg[0], reg[1])
        for reg in RegexPatternCpp.convert_c17_to_c14:
            body = self.replace(body, reg)
        return body



# -*- coding: utf-8 -*-
from tests.test_serializer.py.gen.DataWrapper import DataWrapper
from tests.test_serializer.py.gen.IntrusivePtr import IntrusivePtr
import xml.etree.ElementTree as ET


class SerializerXml(object):
    def __init__(self, node):
        self.node = node

    def serialize(self, obj, key, default_value=None):
        if isinstance(obj, DataWrapper):
            self.serialize_attr(obj.name, key, default_value)
        elif isinstance(obj, IntrusivePtr):
            child = self.add_child(key)
            child.serialize_attr(obj.get_type(), 'type', default_value)
            obj.serialize_xml(child)
        elif hasattr(obj, 'serialize_xml'):
            obj.serialize_xml(self.add_child(key))
        elif isinstance(obj, list):
            self.serialize_list(obj, key)
        elif isinstance(obj, dict):
            self.serialize_dict(obj, key)
        else:
            self.serialize_attr(obj, key, default_value)

    def serialize_attr(self, obj, key, default_value):
        if obj != default_value:
            if obj is True:
                self.node.attrib[key] = 'true'
            elif obj is False:
                if default_value is not None:
                    self.node.attrib[key] = 'false'
            else:
                self.node.attrib[key] = str(obj)

    def serialize_dict(self, obj, key):
        node = SerializerXml(self.node) if not key else self.add_child(key)
        for k, v in obj.items():
            item = node.add_child('pair')
            item.serialize_key(k)
            item.serialize_value(v)

    def serialize_list(self, obj, key):
        node = SerializerXml(self.node) if not key else self.add_child(key)
        for item in obj:
            node.add_child('item').serialize_list_item(item)

    def add_child(self, key):
        node = ET.SubElement(self.node, key) if key else self.node
        return SerializerXml(node)

    def serialize_key(self, k):
        self.serialize(k, 'key')

    def serialize_value(self, v):
        self.serialize(v, 'value')

    def serialize_list_item(self, obj):
        if isinstance(obj, DataWrapper):
            self.serialize_attr(obj.name, 'value', '')
        elif isinstance(obj, IntrusivePtr):
            self.node.tag = obj.get_type()
            obj.serialize_xml(self)
        elif hasattr(obj, 'serialize_xml'):
            self.serialize(obj, '')
        elif isinstance(obj, list):
            self.serialize_list(obj, '')
        elif isinstance(obj, dict):
            self.serialize_dict(obj, '')
        else:
            self.serialize_attr(obj, 'value', None)



from .cpp_extension import FILES_DICT, FACTORY_REGISTRATION
from ..base.writer_base import WriterBase
from ..base.model import SerializeFormat
from ..core.class_ import Class


class GeneratorPredefinedFiles(object):

    def __init__(self):
        pass

    @staticmethod
    def get_namespace():
        return 'mg'

    def generate(self, model):
        writer = WriterBase('')
        writer.model = model
        self.generate_config_files(model)
        self.generate_base_enum_class(model)
        for pair in FILES_DICT:
            filename = pair[0]
            if 'intrusive_ptr' in filename and not model.generate_intrusive:
                continue
            if 'Factory' in filename and not model.generate_factory:
                continue
            content = pair[1]
            content = content.replace('@{namespace}', self.get_namespace())
            content = content.replace('@{namespace_upper}', self.get_namespace().upper())
            content = writer.prepare_file(content)
            filename = filename.replace('@{namespace}', self.get_namespace())
            if 'Factory' in filename:
                registration = FACTORY_REGISTRATION if model.auto_registration else ''
                content = content.replace('@{registration}', registration)
            model.add_file(None, filename, content)

    def generate_config_files(self, model):
        pattern = '#ifndef __{0}_Config_h__\n#define __{0}_Config_h__\n\n{1}\n\n#endif //#ifndef __{0}_Config_h__'
        configs = list()

        for serialize_format, format_string in SerializeFormat.get_all():
            support = 'true' if model.serialize_formats & serialize_format != 0 else 'false'
            configs.append('#define SUPPORT_{}_PROTOCOL {}'.format(format_string.upper(), support))

        model.add_file(None, 'config.h', pattern.format(self.get_namespace(), '\n'.join(configs)))

    def generate_base_enum_class(self, model):
        base_enum = Class()
        base_enum.name = 'BaseEnum'
        base_enum.type = 'class'
        base_enum.auto_generated = False

        model.add_class(base_enum)



from os.path import isfile
import yaml
from mlc_tools.utils.fileutils import normalize_path


class Feature(object):
    GENERATE_TESTS = 'tests'
    INTRUSIVE_PTR = 'intrusive_ptr'
    FACTORY = 'factory'
    PHP_VALIDATE = 'php_validate'
    AUTO_REGISTRATION = 'auto_registration'


class ProjectConfig(object):
    def __init__(self):
        self.has_config = False
        self.root = None
        self.arguments = None
        self.name = None
        self.binary_type = 'app'
        self.serialize_format = None
        self.src_directory = None
        self.src_directory_add = None
        self.data_directory = None
        self.build_directory = None
        self.third_party_source_url = None
        self.third_party_release = None
        self.generate_sources_directory = None
        self.lang = None
        self.serialize_format = None
        self.set_defaults('')
        self.features = {}

    def set_defaults(self, root, arguments=None):
        self.root = root
        self.arguments = arguments
        self.name = 'app'
        self.serialize_format = 'xml'
        self.src_directory = normalize_path(root + 'src')
        self.data_directory = normalize_path(root + 'data')
        self.build_directory = normalize_path(root + 'build/' + arguments.mode if arguments else 'debug')
        self.generate_sources_directory = normalize_path(self.build_directory + 'gen')
        self.third_party_source_url = 'https://github.com/mlc-tools/third_party.git'
        self.third_party_release = 'master'
        self.lang = 'cpp'
        self.serialize_format = 'xml'
        self.features = {}

    def parse(self, root, arguments):
        self.set_defaults(root, arguments)
        root = normalize_path(root)
        config_file = root + arguments.config
        if not isfile(config_file):
            return
        with open(config_file, encoding='utf-8') as stream:
            try:
                data = yaml.safe_load(stream)
                self._parse_dict(data)
            except yaml.YAMLError as exception:
                print(exception)
                raise RuntimeError('Cannot parse project.yaml file')
            self.has_config = True

    def _parse_dict(self, dictionary):
        try:
            self.name = dictionary.get('project')
            self.serialize_format = dictionary.get('serialize_format', self.serialize_format)
            self.binary_type = dictionary.get('binary_type', self.binary_type)
            self.lang = dictionary.get('lang', self.lang)
            if 'src_directory' in dictionary:
                self.src_directory = normalize_path(self.root + dictionary.get('src_directory'))
            if 'src_directory_add' in dictionary:
                self.src_directory_add = normalize_path(self.root + dictionary.get('src_directory_add'))
            if 'data_directory' in dictionary:
                self.data_directory = normalize_path(self.root + dictionary.get('data_directory'))
            if 'build_directory' in dictionary:
                self.build_directory = normalize_path(self.root + dictionary.get('build_directory'))
                self.build_directory += self.arguments.mode if self.arguments else 'debug'
                self.build_directory = normalize_path(self.build_directory)
            if 'generate_sources_directory' in dictionary:
                self.generate_sources_directory = normalize_path(self.root + dictionary.get('generate_sources_directory'))

            if 'features' in dictionary:
                features = dictionary['features']
                features = [x.strip() for x in features.split(' ')]
                for feature in features:
                    self.features[feature] = True

        except KeyError as error:
            print('Project data has not key')
            print(error)
            raise RuntimeError('Cannot parse project data')
        self._validate_values('serialize_format')
        self._validate_values('lang')
        self._validate_values('binary_type')

    def _validate_values(self, name):
        values = {
            'serialize_format': ['xml', 'json'],
            'lang': ['cpp', 'py', 'php', 'js'],
            'binary_type': ['app', 'lib', ''],
        }
        value = self.__getattribute__(name)
        if name in values and value not in values[name]:
            raise RuntimeError(f'Incorrect value of field:\n  {name}: {value}\n  Can be one of {values[name]}')



from ..base.generator_operator_equals import GeneratorOperatorEqualsBase


class GeneratorOperatorEquals(GeneratorOperatorEqualsBase):

    def __init__(self):
        GeneratorOperatorEqualsBase.__init__(self)

    def get_equal_method_name(self):
        return '__eq__'

    def get_not_equal_method_name(self):
        return '__ne__'

    def get_compare_method_pattern(self, cls, member):
        line = 'result = result && this->{0} == rhs.{0};'
        line_ptr = \
            'result = result && ((this->{0} == nullptr && rhs.{0} == nullptr) || ' \
            '(this->{0} != nullptr && rhs.{0} != nullptr && this->{0} == rhs.{0}));'
        return line_ptr if member.is_pointer or member.is_link else line

    def get_not_equal_method_operation(self):
        return 'return !(this == rhs);'

    def get_call_superclass_equal(self):
        return '{}.{}(self, rhs);'



from ..base.generator_data_storage_base import GeneratorDataStorageBase
from ..base.generator_data_storage_base import get_data_list_name, get_data_name
from ..core.object import Object, Objects
from ..core.function import Function


class GeneratorDataStorage(GeneratorDataStorageBase):

    def __init__(self):
        GeneratorDataStorageBase.__init__(self)

    def is_need_create_static_instance(self):
        return False

    def get_shared_method_body(self):
        return '''static DataStorage instance;
                  return instance;'''

    def create_getters(self, classes):
        method = Function()
        method.name = 'get'
        method.args.append(['name', Objects.STRING])
        method.return_type = Object()
        method.return_type.type = 'template <class T> const T*'
        method.is_const = True
        method.is_template = True
        self.functions.append(method)
        self.generate_implementations(classes, method)
        return method

    def generate_implementations(self, classes, getter):
        implementations = ''
        classes = self.get_storage_classes()
        for class_ in classes:
            impl = '''
            template<>const {type}* DataStorage::get(const std::string& name) const
            {{
                if(name.empty())
                {{
                    return nullptr;
                }}
                if(_loaded)
                {{
                    auto iter = {name}.find(name);
                    if(iter == {name}.end())
                    {{
                        std::cout << "Cannot find data with name [" << name << "] in DataStorage::{name}" << std::endl; 
                    }}
                    return iter != {name}.end() ? &iter->second : nullptr;
                }}
                return &const_cast<DataStorage*>(this)->{name}[name];
            }}
            '''
            name = get_data_list_name(get_data_name(class_.name))
            implementations += impl.format(name=name, type=class_.name)
        getter.specific_implementations = implementations

    def get_storage_classes(self):
        return [x for x in self.model.classes if x.is_storage and x.side in [self.model.side, 'both']]

    def get_initialize_function_json_body(self):
        collect_objects = ''
        classes = self.get_storage_classes()
        for class_ in classes:
            data_name = get_data_name(class_.name)
            map_name = get_data_list_name(data_name)
            collect_objects += f'''
            auto {data_name} = json["{map_name}"];
            for(auto& node : {data_name})
            {{
                auto name = node["key"].asString();
                non_const_this->{map_name}.emplace(name, {class_.name}());
            }} 
            '''
        return f'''Json::Value json;
        Json::Reader reader;
        reader.parse(content, json);
        auto non_const_this = const_cast<DataStorage*>(this);
        {collect_objects}
        DeserializerJson deserializer(json);
        non_const_this->_loaded = true;
        non_const_this->deserialize_json(deserializer);
        '''

    def get_initialize_function_xml_body(self):
        collect_objects = ''
        classes = self.get_storage_classes()
        for class_ in classes:
            data_name = get_data_name(class_.name)
            map_name = get_data_list_name(data_name)
            collect_objects += f'''
            for(auto& node : root.child("{map_name}"))
            {{
                auto name = node.attribute("key").as_string();
                non_const_this->{map_name}.emplace(name, {class_.name}());
            }}
            '''
        return f'''pugi::xml_document doc;
        doc.load_string(content.c_str());
        auto non_const_this = const_cast<DataStorage*>(this);
        if(doc.root() != nullptr)
        {{
            pugi::xml_node root = doc.root().first_child();
            {collect_objects}
            DeserializerXml deserializer(root);
            non_const_this->_loaded = true;
            non_const_this->deserialize_xml(deserializer);
        }}'''



from __future__ import print_function

import collections
import os
import sys
import subprocess
from inspect import getmembers

from .base.circular_reference import CircularReference
from .utils.error import Log, Error
from .utils import fileutils
from .base import Parser
from .base import Linker
from .base import Validator
from .base import DataParser
from .base import Language
from .base.model import Model, SerializeFormat


class Mlc(object):

    def __init__(self, **kwargs):
        self.model = Model()

        Log.use_colors = kwargs.get('use_colors', False)
        Log.disable_logs = kwargs.get('disable_logs', False)
        self._parse_kwargs(**kwargs)

    def _parse_kwargs(self, **kwargs):
        self.model.configs_directory = kwargs.get('configs_directory', self.model.configs_directory)
        self.model.out_directory = kwargs.get('out_directory', self.model.out_directory)
        self.model.data_directory = kwargs.get('data_directory', self.model.data_directory)
        self.model.out_data_directory = kwargs.get('out_data_directory',
                                                   self.model.out_data_directory)
        self.model.language = kwargs.get('language', self.model.language)
        self.model.only_data = kwargs.get('only_data', self.model.only_data)
        self.model.namespace = kwargs.get('namespace', self.model.namespace)
        self.model.side = kwargs.get('side', self.model.side)
        self.model.php_validate = kwargs.get('php_validate', self.model.php_validate)
        self.model.validate_allow_different_virtual_method = kwargs.get('validate_allow_different_virtual_method',
                                                             self.model.validate_allow_different_virtual_method)
        self.model.test_script = kwargs.get('test_script', self.model.test_script)
        self.model.test_script_args = kwargs.get('test_script_args', self.model.test_script_args)
        self.model.generate_tests = kwargs.get('generate_tests', self.model.generate_tests)
        self.model.generate_intrusive = kwargs.get('generate_intrusive',
                                                   self.model.generate_intrusive)
        self.model.generate_factory = kwargs.get('generate_factory', self.model.generate_factory)
        self.model.out_directory = fileutils.normalize_path(self.model.out_directory)
        self.model.out_data_directory = fileutils.normalize_path(self.model.out_data_directory)
        self.model.join_to_one_file = kwargs.get('join_to_one_file', self.model.join_to_one_file)
        self.model.auto_registration = kwargs.get('auto_registration', self.model.auto_registration)
        self.model.generate_ref_counter = kwargs.get('generate_ref_counter', self.model.generate_ref_counter)
        self.model.user_includes = kwargs.get('user_includes', self.model.user_includes)
        if 'add_config' in kwargs and kwargs.get('add_config') is not None:
            directory = fileutils.normalize_path(kwargs.get('add_config'))
            self.model.additional_config_directories.append(directory)
        if 'add_data' in kwargs:
            directory = fileutils.normalize_path(kwargs.get('add_data'))
            self.model.additional_data_directories.append(directory)

        if 'formats' in kwargs:
            formats = kwargs['formats'].split(',')
            self.model.serialize_formats = 0
            for serialize_format, string_format in SerializeFormat.get_all():
                if string_format in formats:
                    self.model.serialize_formats = self.model.serialize_formats | serialize_format

    def add_config_directories(self, directory):
        self.model.additional_config_directories.append(fileutils.normalize_path(directory))

    def add_data_directories(self, directory):
        self.model.additional_data_directories.append(fileutils.normalize_path(directory))

    def set_filter_code(self, func):
        self.model.filter_code = func

    def set_filter_data(self, func):
        self.model.filter_data = func

    def generate(self, **kwargs):
        self.model.clear_data()
        self._parse_kwargs(**kwargs)

        def get_config_files():
            all_directories = [self.model.configs_directory]
            all_directories.extend(self.model.additional_config_directories)
            files = []
            for directory in all_directories:
                directory = fileutils.normalize_path(directory)
                add_files = fileutils.get_files_list(directory)
                add_files = [directory + f for f in add_files]
                files.extend(add_files)

            result_files = []
            for path in files:
                if path.endswith('.mlc'):
                    if isinstance(self.model.filter_code, collections.abc.Callable) and not self.model.filter_code(path):
                        continue
                    result_files.append(path)
            return result_files

        all_files = get_config_files()
        parser = Parser(self.model)
        parser.parse_files(all_files)
        self.model.parser = parser

        language = Language(self.model.language, self.model)
        language.get_generator().generate(self.model)

        Linker().link(self.model)
        Validator().validate(self.model)

        self.run_user_generator()
        if language.registrar:
            language.registrar.generate(self.model)

        CircularReference(self.model).find()

        language.get_translator().translate(self.model)
        language.get_serializer().generate_methods(self.model)
        language.get_writer().save(self.model)
        language.save_plugin.save_files(self.model.join_to_one_file)

    def generate_data(self, **kwargs):
        self._parse_kwargs(**kwargs)

        classes = []
        for class_ in self.model.classes:
            if class_.is_storage:
                classes.append(class_)
        for class_ in self.model.classes_for_data:
            if class_.is_storage:
                classes.append(class_)
        data_parser = DataParser(classes, self.model.data_directory, self.model.filter_data)
        data_parser.parse(self.model.additional_data_directories)
        data_parser.flush(self.model.out_data_directory)

    def run_test(self, **kwargs):
        self._parse_kwargs(**kwargs)

        if self.model.test_script and os.path.isfile(self.model.test_script):
            python = 'python3' if sys.version_info[0] == 3 else 'python'
            command = '{} {} {}'.format(python, self.model.test_script, self.model.test_script_args)
            Log.message('Run test (%s):' % command)

            if subprocess.call([python, self.model.test_script, self.model.test_script_args]):
                Error.exit(Error.TESTS_FAILED)
        if not os.path.isfile(self.model.test_script):
            Log.error('Test script (%s) not founded' % self.model.test_script)
            sys.exit(1)

    def set_user_generator(self, generator):
        self.model.custom_generator = generator

    def run_user_generator(self):
        if self.model.custom_generator:
            self.model.custom_generator.execute(self.model)



from mlc_tools.base.model import Model, SerializeFormat
from mlc_tools.core.function import Function
from mlc_tools.core.object import Objects
from ..base.generator_data_storage_base import GeneratorDataStorageBase


class GeneratorDataStorage(GeneratorDataStorageBase):

    def __init__(self):
        GeneratorDataStorageBase.__init__(self)

    def get_shared_method_body(self):
        return '''        if not DataStorage.__instance:
    DataStorage.__instance = DataStorage()
return DataStorage.__instance'''

    def get_pattern_getter(self):
        return '''        if name == '':
            return None
        if not self._loaded and name not in self.{map}:
            from .{type} import {type}
            self.{map}[name] = {type}()
            self.{map}[name].name = name
        from .DataWrapper import DataWrapper
        return DataWrapper(self.{map}[name] if name in self.{map} else None)'''

    def get_initialize_function_json_body(self):
        return '''        js = json.loads(content)
        deserializer = DeserializerJson(js)
        self.deserialize_json(deserializer)
        self._loaded = True'''

    def get_initialize_function_xml_body(self):
        return '''        root = ET.fromstring(content)
        deserializer = DeserializerXml(root)
        self.deserialize_xml(deserializer)
        self._loaded = True
        '''

    def generate(self, model: Model):
        GeneratorDataStorageBase.generate(self, model)
        if model.serialize_formats & SerializeFormat.xml:
            self.add_deserialize_xml(model)
        if model.serialize_formats & SerializeFormat.json:
            self.add_deserialize_json(model)

    def add_deserialize_json(self, model):
        method = self.add_deserialize_method('deserialize_json')
        for key, obj in self.data_members.items():
            getter = 'get' + key
            map_name = obj.name
            method.operations.append(f"{map_name} = deserializer.json.get('{map_name}', None)")
            method.operations.append(f"if {map_name}:")
            method.operations.append(f"    for o in {map_name}:")
            method.operations.append(f"        self.{getter}(o['key'])")
            method.operations.append(f"    for o in {map_name}:")
            method.operations.append(f"        data = self.{getter}(o['key'])")
            method.operations.append(f"        data.deserialize_json(DeserializerJson(o['value']))")

    def add_deserialize_xml(self, model):
        method = self.add_deserialize_method('deserialize_xml')
        for key, obj in self.data_members.items():
            getter = 'get' + key
            map_name = obj.name
            method.operations.append(f"{map_name} = deserializer.node.find('{map_name}')")
            method.operations.append(f"if {map_name}:")
            method.operations.append(f"    for o in {map_name}:")
            method.operations.append(f"        self.{getter}(o.attrib['key'])")
            method.operations.append(f"    for o in {map_name}:")
            method.operations.append(f"        data = self.{getter}(o.attrib['key'])")
            method.operations.append(f"        if data:")
            method.operations.append(f"            deserializer_data = DeserializerXml(o.find('value'))")
            method.operations.append(f"            data.deserialize_xml(deserializer_data)")

    def add_deserialize_method(self, name):
        method = Function()
        method.name = name
        method.translated = True
        method.return_type = Objects.VOID
        method.is_virtual = False
        method.is_const = False
        method.operations.append('        pass')
        method.args.append(['deserializer', self])
        self.functions.append(method)
        return method





from ..base.save_plugin_base import SavePluginBase


class SavePlugin(SavePluginBase):

    def __init__(self, model):
        SavePluginBase.__init__(self, model)

    def save_files(self, combine_to_one=True):
        SavePluginBase.save_files(self, True)

    def _create_combine_file_header(self):
        header = ''
        file_name = 'mg.js'
        return header, file_name

    def _finalize_combine_file(self, content):
        return content

    def _remove_includes(self, file_content):
        return file_content



FACTORY = '''
class Factory
{
}
Factory.build = function (type)
{
    return eval("new " + type + "()");
};

'''

COMMON = '''
create_command_from_json = function (payload)
{
    let json = JSON.parse(payload);
    for(let key in json)
    {
        let command = Factory.build(key);
        if(command)
        {
            command.deserialize_json(json[key]);
        }
        return command;
    }
};

serialize_command_to_json = function (command)
{
    let json = {};
    json[command.get_type()] = {};
    command.serialize_json(json[command.get_type()]);
    return JSON.stringify(json);
};

clone_object = function (obj)
{
    let payload = serialize_command_to_json(obj);
    return create_command_from_json(payload);
};
'''


from .constants import FACTORY
from .writer import Writer


class GeneratorFactory(object):

    def __init__(self):
        pass

    @staticmethod
    def generate(model):
        writer = Writer('')
        writer.model = model
        content = writer.prepare_file(FACTORY)
        model.add_file(None, 'Factory.js', content)



from ..core.class_ import Class


PREDEFINED = '''
class @{name}
{
    constructor()
    {
        this.listeners = [];
        this.objects = [];
    }

    add(object, callback)
    {
        this.objects.push(object);
        this.listeners.push(callback);
    }
    remove(object)
    {
        let index = this.objects.indexOf(object);
        if(index != -1)
        {
            this.listeners.splice(index, 1);
            this.objects.splice(index, 1);
        }
    }
    notify()
    {
        for (let index in this.listeners)
        {
            let object = this.objects[index];
            let callback = this.listeners[index];
            callback.apply(object, arguments);
        }
    }
}

'''


class GeneratorObserver(object):

    def __init__(self):
        pass

    @staticmethod
    def get_mock():
        cls = Class()
        cls.name = GeneratorObserver.get_observable_name()
        cls.type = 'class'
        cls.auto_generated = False
        return cls

    @staticmethod
    def get_observable_name():
        return 'Observable'

    @staticmethod
    def generate(model):
        text = PREDEFINED
        filename = GeneratorObserver.get_observable_name() + '.js'
        text = text.replace('@{name}', GeneratorObserver.get_observable_name())

        mock = GeneratorObserver.get_mock()
        model.add_file(mock, filename, text)
        model.add_class(mock)



import re

from .function import Function
from .object import Object
from .modifiers import Modifiers
from ..utils.error import Error


class Class(Object):

    OPERATOR_MATCH = re.compile(r'operator\b.+')

    def __init__(self, name=''):
        Object.__init__(self)
        self.superclasses = []
        self.subclasses = []
        self.constructor = None
        self.members = []
        self.functions = []
        self.user_includes = set()
        self.is_abstract = False
        self.is_serialized = False
        self.is_visitor = False
        self.is_storage = False
        self.is_numeric = False
        self.is_test = False
        self.is_inline = False
        self.is_virtual = False
        self.generate_set_function = False
        self.type = 'class'
        self.group = ''
        self.side = 'both'
        self.auto_generated = True
        self.inner_classes = []
        self._linked = False
        self.prefer_use_forward_declarations = False

        self.name = name

    def has_method_with_name(self, name):
        for method in self.functions:
            if method.name == name:
                return True
        return False

    def has_member_with_name(self, name):
        for member in self.members:
            if member.name == name:
                return True
        return False

    def get_method_with_name(self, name) -> Function or None:
        for method in self.functions:
            if method.name == name:
                return method
        return None

    def has_abstract_method(self):
        for method in self.functions:
            if method.is_abstract:
                return True
        return False

    def has_virtual_table(self):
        result = False
        result = result or self.is_virtual
        result = result or self.superclasses
        result = result or self.subclasses
        result = result or self.has_abstract_method()
        return result

    def on_linked(self, model):
        if self._linked:
            return

        for func in self.functions:
            func.is_virtual = self.is_virtual or \
                func.is_virtual or \
                func.is_abstract or \
                self.has_function_in_subclasses(func)

            # if 'operator' in func.name:
            if Class.OPERATOR_MATCH.findall(func.name):
                func.is_virtual = False
            func.is_virtual = func.is_virtual or \
                self.has_function_in_superclasses(func)

        if not self.is_abstract:
            for method in self.functions:
                if method.is_abstract:
                    self.is_abstract = True
                    method.is_virtual = True
                    break
            for superclass in self.superclasses:
                superclass.on_linked(model)

        self._linked = True

    @staticmethod
    def __has_function(list_of_classes, method):
        for cls in list_of_classes:
            for func in cls.functions:
                equal = True
                equal = equal and func.name == method.name
                equal = equal and func.return_type.type == method.return_type.type and 'operator' not in func.name
                if equal:
                    func.is_virtual = True
                    return True
        # for cls in list_of_classes:
        #     if cls.has_function_in_subclasses(method):
        #         return True
        return False


    def has_function_in_subclasses(self, method):
        return Class.__has_function(self.subclasses, method)

    def has_function_in_superclasses(self, method):
        return Class.__has_function(self.superclasses, method)

    def find_modifiers(self, string):
        self.is_abstract = self.is_abstract or Modifiers.abstract in string
        self.is_serialized = self.is_serialized or Modifiers.serialized in string
        self.is_visitor = self.is_visitor or Modifiers.visitor in string
        self.is_storage = self.is_storage or Modifiers.storage in string
        self.is_numeric = self.is_numeric or Modifiers.numeric in string
        self.is_test = self.is_test or Modifiers.test in string
        self.is_inline = self.is_inline or Modifiers.inline in string
        self.is_virtual = self.is_virtual or Modifiers.virtual in string
        self.generate_set_function = self.generate_set_function or Modifiers.set_function in string
        self.prefer_use_forward_declarations = self.prefer_use_forward_declarations or Modifiers.prefer_use_forward_declarations in string

        if Modifiers.server in string:
            self.side = Modifiers.side_server
        if Modifiers.client in string:
            self.side = Modifiers.side_client

        string = string.replace(Modifiers.server, '')
        string = string.replace(Modifiers.client, '')
        string = string.replace(Modifiers.abstract, '')
        string = string.replace(Modifiers.serialized, '')
        string = string.replace(Modifiers.visitor, '')
        string = string.replace(Modifiers.storage, '')
        string = string.replace(Modifiers.set_function, '')
        string = string.replace(Modifiers.numeric, '')
        string = string.replace(Modifiers.test, '')
        string = string.replace(Modifiers.inline, '')
        string = string.replace(Modifiers.virtual, '')
        string = string.replace(Modifiers.prefer_use_forward_declarations, '')

        return string

    def generate_constructor(self):
        counter = 0
        for method in self.functions:
            if method.name == 'constructor':
                self.constructor = method
                if counter > 0:
                    Error.exit(Error.CLASS_HAVE_MORE_THAN_ONE_CONSTRUCTOR, self.name)

        if self.constructor and self.constructor.is_generate:
            for member in self.members:
                arg = Object()
                arg.type = member.type
                arg.set_default_initial_value()
                self.constructor.args.append([member.name, arg])
                self.constructor.operations.append('this->{0} = {0};'.format(member.name))


import unittest
from mlc_tools.base.model import Model
from mlc_tools.core.class_ import Class

class TestModel(unittest.TestCase):

    def setUp(self):
        self.model = Model()

    def test_empty_copy(self):
        new_model = self.model.empty_copy()
        self.assertNotEqual(id(new_model), id(self.model))
        self.assertEqual(new_model.classes, [])
        self.assertEqual(new_model.classes_for_data, [])
        self.assertEqual(new_model.objects, [])
        self.assertEqual(new_model.functions, [])
        self.assertEqual(new_model.classes_dict, {})

    def test_clear_data(self):
        self.model.parser = 'some_parser'
        self.model.classes = ['some_class']
        self.model.classes_for_data = ['some_class_for_data']
        self.model.objects = ['some_object']
        self.model.functions = ['some_function']
        self.model.classes_dict = {'some_class': Class('some_class')}

        self.model.clear_data()

        self.assertIsNone(self.model.parser)
        self.assertEqual(self.model.classes, [])
        self.assertEqual(self.model.classes_for_data, [])
        self.assertEqual(self.model.objects, [])
        self.assertEqual(self.model.functions, [])
        self.assertEqual(self.model.classes_dict, {})
        self.assertIsNone(self.model.out_dict)
        self.assertEqual(self.model.files, [])
        self.assertEqual(self.model.created_files, [])

    def test_add_class(self):
        cls = Class('TestClass')
        self.model.add_class(cls)
        self.assertEqual(len(self.model.classes), 1)
        self.assertEqual(self.model.classes[0], cls)
        self.assertEqual(len(self.model.classes_dict), 1)
        self.assertEqual(self.model.classes_dict['TestClass'], cls)

    def test_add_classes(self):
        classes = [Class('TestClass1'), Class('TestClass2')]
        self.model.add_classes(classes)
        self.assertEqual(len(self.model.classes), 2)
        self.assertEqual(self.model.classes[0], classes[0])
        self.assertEqual(self.model.classes[1], classes[1])
        self.assertEqual(len(self.model.classes_dict), 2)
        self.assertEqual(self.model.classes_dict['TestClass1'], classes[0])
        self.assertEqual(self.model.classes_dict['TestClass2'], classes[1])

    def test_get_class(self):
        cls = Class('TestClass')
        self.model.add_class(cls)
        self.assertEqual(self.model.get_class('TestClass'), cls)

    def test_has_class(self):
        cls = Class('TestClass')
        self.model.add_class(cls)
        self.assertTrue(self.model.has_class('TestClass'))
        self.assertFalse(self.model.has_class('NonExistentClass'))

    def test_is_side(self):
        self.model.side = 'server'
        self.assertTrue(self.model.is_side('server'))
        self.assertFalse(self.model.is_side('client'))
        self.assertTrue(self.model.is_side('both'))

    def test_is_lang(self):
        self.model.language = 'py'
        self.assertTrue(self.model.is_lang('py'))
        self.assertFalse(self.model.is_lang('cpp'))
        self.assertTrue(self.model.is_lang(None))



from copy import deepcopy
from ..core.class_ import Class
from ..core.object import Object, Objects, AccessSpecifier
from ..core.function import Function
from ..utils.error import Error


class Linker(object):

    def __init__(self):
        pass

    def link(self, model):
        self.add_functions_to_classes(model)
        self.add_members_to_classes(model)
        self.convert_superclasses(model)
        self.convert_templates(model)
        self.generate_inline_classes(model)
        self.add_get_type_method(model)
        for cls in model.classes:
            cls.on_linked(model)

    @staticmethod
    def add_get_type_method(model):
        for cls in model.classes:
            if cls.type == 'class' and cls.auto_generated:
                Linker.add_get_type_function(cls)

    @staticmethod
    def generate_inline_classes(model):
        for cls in model.classes:
            Linker._generate_inline_classes(cls)

    @staticmethod
    def _generate_inline_classes(class_):
        if not class_.superclasses:
            return
        for superclass in class_.superclasses:
            assert isinstance(superclass, Class)
            if superclass.__class__ != Class:
                Error.exit(Error.INTERNAL_ERROR)

            if not superclass.is_inline:
                continue
            for obj in superclass.members:
                copy = deepcopy(obj)
                class_.members.append(copy)
            for func in superclass.functions:
                disallow = False
                disallow = disallow or func.name == 'operator =='
                disallow = disallow or func.name == 'operator !='
                if disallow:
                    continue
                copy = deepcopy(func)
                class_.functions.append(copy)

            superclass.subclasses.remove(class_)

        class_.superclasses = [i for i in class_.superclasses if not i.is_inline]

    @staticmethod
    def convert_templates(model):
        for cls in model.classes:
            for member in cls.members:
                Linker._convert_templates(model, member)

    @staticmethod
    def _convert_templates(model, member):
        args = []
        for arg in member.template_args:
            if arg.__class__ == Object:
                args.append(arg)
            else:
                args.append(Linker.get_object_type(model, arg))
                if args[-1].__class__ == Object:
                    Linker._convert_templates(model, args[-1])
        member.template_args = args

    @staticmethod
    def get_object_type(model, type_name):
        cls = model.get_class(type_name)
        obj = Object()
        if cls is None:
            obj.type = obj.find_modifiers(type_name)
            obj.parse_type()
        else:
            obj.type = cls.name
        obj.name = ""
        return obj

    @staticmethod
    def convert_superclasses(model):
        for cls in model.classes:
            superclasses = []
            for name in cls.superclasses:
                superclass = model.get_class(name)
                if superclass is None:
                    Error.exit(Error.UNKNOWN_SUPERCLASS, cls.name, name)
                superclasses.append(superclass)
                superclass.subclasses.append(cls)
            cls.superclasses = superclasses

    @staticmethod
    def add_members_to_classes(model):
        for object_ in model.objects:
            parts = object_.name.split('::')
            if len(parts) == 2:
                class_name = parts[0]
                object_name = parts[1]
                class_ = model.get_class(class_name)
                if class_ is None:
                    Error.exit(Error.CANNOT_FIND_CLASS_FOR_OBJECT, class_name, object_.name)
                object_.name = object_name
                class_.members.append(object_)
                model.objects.remove(object_)

    @staticmethod
    def add_functions_to_classes(model):
        for method in model.functions:
            parts = method.name.split('::')
            if len(parts) == 2:
                class_name = parts[0]
                function_name = parts[1]
                class_ = model.get_class(class_name)
                if class_ is None:
                    Error.exit(Error.CANNOT_FIND_CLASS_FOR_METHOD, class_name, method.name)
                method.name = function_name
                class_.functions.append(method)
        model.functions = []

    @staticmethod
    def add_get_type_function(class_):
        member = Object()
        member.is_static = True
        member.is_const = True
        member.type = 'string'
        member.name = 'TYPE'
        member.initial_value = '"{}"'.format(class_.name)
        member.access = AccessSpecifier.public
        class_.members.append(member)

        method = Function()
        method.name = 'get_type'
        method.return_type = Objects.STRING
        method.is_const = True
        method.operations.append('return {}::TYPE;'.format(class_.name))
        class_.functions.append(method)
        method.is_virtual = class_.is_virtual or class_.superclasses or class_.subclasses



# -*- coding: utf-8 -*-
import xml.etree.ElementTree as ET
import json
from .common import *
from .BaseEnum import BaseEnum


class TestEnum(BaseEnum):
    value1 = "value1"
    value2 = "value2"
    __slots__ = []

    def __init__(self):
        BaseEnum.__init__(self)

    def __hash__(self):
        return id(self)



import re
from .regex import RegexPatternPython
from ..base.translator_base import TranslatorBase
from ..core.class_ import Class
from ..core.function import Function
from ..core.object import Objects, Object


class Translator(TranslatorBase):

    def __init__(self):
        TranslatorBase.__init__(self)

    def translate_function(self, cls, method, model):
        if not method.translated:
            body = '\n'.join(method.operations)
            body = self.translate_function_body(cls, method, body, model, method.args)
            method.body = body
        else:
            if method.operations:
                method.body = '\n        '.join(method.operations)
            else:
                method.body = 'pass'

    def translate_function_body(self, cls, method, body, model, args):
        if not body:
            body = 'pass'
        body = self.replace_by_regex(method, body, model, args)
        body = Translator.convert_braces_to_tabs(body)
        body = Translator.remove_double_eol(body)
        body = Translator.add_imports(cls, body, model)
        body = Translator.convert_braces(body)
        return body

    @staticmethod
    def add_imports(cls_owner, func, model):
        if not func:
            return func
        imports = []
        if RegexPatternPython.FACTORY.search(func):
            imports.append('from .Factory import Factory')
        for cls in model.classes:
            if cls.name in func:
                need = cls.name is not cls_owner.name
                if not need:
                    continue
                pattern = None
                if cls.name not in RegexPatternPython.regs_class_names:
                    pattern = re.compile(r'\b{}\b'.format(cls.name))
                    RegexPatternPython.regs_class_names[cls.name] = pattern
                pattern = pattern or RegexPatternPython.regs_class_names[cls.name]
                need = need and pattern.search(func) is not None
                if need:
                    imports.append('from .{0} import {0}'.format(cls.name))
        if 'math.' in func:
            imports.append('import math')
        if 'random.' in func:
            imports.append('import random')
        if imports:
            imports = Translator.get_tabs(2) + '\n{}'.format(Translator.get_tabs(2)).join(imports)
            func = '{}\n{}'.format(imports, func)
        return func

    @staticmethod
    def remove_double_eol(func):
        if not func:
            return func
        func = func.replace('\n    \n', '\n')
        func = func.replace('\n        \n', '\n')
        func = func.replace('\n            \n', '\n')
        func = func.replace('\n                \n', '\n')
        func = func.replace('\n                    \n', '\n')
        func = func.replace('\n                        \n', '\n')
        return func

    @staticmethod
    def convert_braces_to_tabs(func):
        if not func:
            return func
        lines = func.split('\n')
        tabs = 2
        next_tab = False
        for i, line in enumerate(lines):
            if next_tab:
                tabs += 1
            if '{' in line:
                tabs += 1
                line = line.replace('{', '')
            if '}' in line:
                tabs -= 1
                line = line.replace('}', '')
            lines[i] = Translator.get_tabs(tabs) + line.strip()
            if next_tab:
                next_tab = False
                tabs -= 1
            if (line.startswith('for') or line.startswith('if') or line.startswith('else') or line.startswith('elif')
                    or line.startswith('try:') or line.strip().startswith('except')) \
                    and (i < len(lines) - 1 and '{' not in line and '{' not in lines[i + 1]):
                next_tab = True

        func = '\n'.join(lines)
        return func

    @staticmethod
    def convert_braces(body):
        return body.replace('[@[', '{').replace(']@]', '}')

    def replace_by_regex(self, func, body, model, args):
        if not body:
            return body
        for reg in RegexPatternPython.FUNCTION:
            body = self.replace(body, reg)
        for reg in RegexPatternPython.REPLACES:
            body = body.replace(reg[0], reg[1])
        return body

    @staticmethod
    def get_tabs(count):
        return '    ' * count

    def convert_to_enum(self, cls: Class):
        TranslatorBase.convert_to_enum(self, cls)
        cast = cls.members[-1].type

        member = Object()
        member.name = '_value'
        member.type = Objects.STRING
        cls.members.append(member)

        setter = Function()
        setter.name = 'set'
        setter.args.append(['value', Objects.VOID])
        if cast == 'int':
            member.type = Objects.INT
            for index, obj in enumerate(cls.members):
                if obj.name != '_value':
                    setter.operations.append(f'''if(value == "{obj.name}")
                    {{
                        this->_value = {cls.name}::{obj.name};
                        return;
                    }}''')
        else:
            setter.operations.append('this->_value = value;')
        cls.functions.append(setter)

        getter = Function()
        getter.name = 'str'
        if cast == 'int':
            for index, obj in enumerate(cls.members):
                if obj.name != '_value':
                    getter.operations.append(f'''if(this->_value == {cls.name}::{obj.name})
                    {{
                        return "{obj.name}";
                    }}''')
        else:
            getter.operations.append('return this->_value;')
        cls.functions.append(getter)

        def add_binary_method(name, operator):
            method = Function()
            method.name = name
            method.return_type = Objects.BOOL
            method.args.append(['rhs', Objects.VOID])
            method.operations.append(f'''if(isinstance(rhs, {cls.name}))
             {{
                return this->_value {operator} rhs._value;
             }}
             else
             {{
                return this->_value {operator} rhs
             }}''')
            cls.functions.append(method)
        add_binary_method('__eq__', '==')
        add_binary_method('__and__', '&')
        add_binary_method('__or__', '|')
        add_binary_method('__xor__', '^')

